"""
This module defines general-purpose objects, functions and classes.

Functions, classes, etc. defined here should not depend on any other
part of ``stk``. They must be completely self-sufficient.

"""

import rdkit.Chem.AllChem as rdkit
from rdkit.Geometry import Point3D
import numpy as np
import time
from contextlib import contextmanager
import os
import subprocess as sp
import gzip
import re
from collections import deque
import tarfile
from glob import iglob


# Holds the elements Van der Waals radii in Angstroms.
atom_vdw_radii = {
    'Al': 2, 'Sb': 2, 'Ar': 1.88, 'As': 1.85, 'Ba': 2,
    'Be': 2, 'Bi': 2, 'B': 2, 'Br': 1.85, 'Cd': 1.58,
    'Cs': 2, 'Ca': 2, 'C': 1.7, 'Ce': 2, 'Cl': 1.75,
    'Cr': 2, 'Co': 2, 'Cu': 1.4, 'Dy': 2, 'Er': 2,
    'Eu': 2, 'F': 1.47, 'Gd': 2, 'Ga': 1.87, 'Ge': 2,
    'Au': 1.66, 'Hf': 2, 'He': 1.4, 'Ho': 2, 'H': 1.09,
    'In': 1.93, 'I': 1.98, 'Ir': 2, 'Fe': 2, 'Kr': 2.02,
    'La': 2, 'Pb': 2.02, 'Li': 1.82, 'Lu': 2, 'Mg': 1.73,
    'Mn': 2, 'Hg': 1.55, 'Mo': 2, 'Nd': 2, 'Ne': 1.54,
    'Ni': 1.63, 'Nb': 2, 'N': 1.55, 'Os': 2, 'O': 1.52,
    'Pd': 1.63, 'P': 1.8, 'Pt': 1.72, 'K': 2.75, 'Pr': 2,
    'Pa': 2, 'Re': 2, 'Rh': 2, 'Rb': 2, 'Ru': 2, 'Sm': 2,
    'Sc': 2, 'Se': 1.9, 'Si': 2.1, 'Ag': 1.72, 'Na': 2.27,
    'Sr': 2, 'S': 1.8, 'Ta': 2, 'Te': 2.06, 'Tb': 2,
    'Tl': 1.96, 'Th': 2, 'Tm': 2, 'Sn': 2.17, 'Ti': 2,
    'W': 2, 'U': 1.86, 'V': 2, 'Xe': 2.16, 'Yb': 2,
    'Y': 2, 'Zn': 1.29, 'Zr': 2, 'X': 1.0, 'D': 1.0
}

# This dictionary gives easy access to the rdkit bond types.
bond_dict = {'1': rdkit.rdchem.BondType.SINGLE,
             'am': rdkit.rdchem.BondType.SINGLE,
             '2': rdkit.rdchem.BondType.DOUBLE,
             '3': rdkit.rdchem.BondType.TRIPLE,
             'ar': rdkit.rdchem.BondType.AROMATIC}

# A dictionary which matches atomic number to elemental symbols.
periodic_table = {
    1: 'H', 2: 'He', 3: 'Li', 4: 'Be', 5: 'B', 6: 'C',
    7: 'N', 8: 'O', 9: 'F', 10: 'Ne', 11: 'Na', 12: 'Mg',
    13: 'Al', 14: 'Si', 15: 'P', 16: 'S', 17: 'Cl',
    18: 'Ar', 19: 'K', 20: 'Ca', 21: 'Sc', 22: 'Ti',
    23: 'V', 24: 'Cr', 25: 'Mn', 26: 'Fe', 27: 'Co',
    28: 'Ni', 29: 'Cu', 30: 'Zn', 31: 'Ga', 32: 'Ge',
    33: 'As', 34: 'Se', 35: 'Br', 36: 'Kr', 37: 'Rb',
    38: 'Sr', 39: 'Y', 40: 'Zr', 41: 'Nb', 42: 'Mo',
    43: 'Tc', 44: 'Ru', 45: 'Rh', 46: 'Pd', 47: 'Ag',
    48: 'Cd', 49: 'In', 50: 'Sn', 51: 'Sb', 52: 'Te',
    53: 'I', 54: 'Xe', 55: 'Cs', 56: 'Ba', 57: 'La',
    58: 'Ce', 59: 'Pr', 60: 'Nd', 61: 'Pm', 62: 'Sm',
    63: 'Eu', 64: 'Gd', 65: 'Tb', 66: 'Dy', 67: 'Ho',
    68: 'Er', 69: 'Tm', 70: 'Yb', 71: 'Lu', 72: 'Hf',
    73: 'Ta', 74: 'W', 75: 'Re', 76: 'Os', 77: 'Ir',
    78: 'Pt', 79: 'Au', 80: 'Hg', 81: 'Tl', 82: 'Pb',
    83: 'Bi', 84: 'Po', 85: 'At', 86: 'Rn', 87: 'Fr',
    88: 'Ra', 89: 'Ac', 90: 'Th', 91: 'Pa', 92: 'U',
    93: 'Np', 94: 'Pu', 95: 'Am', 96: 'Cm', 97: 'Bk',
    98: 'Cf', 99: 'Es', 100: 'Fm', 101: 'Md', 102: 'No',
    103: 'Lr', 104: 'Rf', 105: 'Db', 106: 'Sg', 107: 'Bh',
    108: 'Hs', 109: 'Mt', 110: 'Ds', 111: 'Rg', 112: 'Cn',
    113: 'Uut', 114: 'Fl', 115: 'Uup', 116: 'Lv',
    117: 'Uus', 118: 'Uuo'
}


class Cell:
    """
    Represents an individual cell in a supercell.

    Parameters
    ----------
    id : :class:`list` of :class:`int`
        A 3 member :class:`list` holding the x, y and z index
        of the cell within the supercell.

    fgs : :class:`dict`
        Maps the fgs in the original unit cell to the
        equivalent fgs in the cell.

    Attributes
    ----------
    id : :class:`numpy.ndarray` of :class:`int`
        A 3 member array holding the x, y and z index
        of the cell within the supercell.

    fgs : :class:`dict`
        Maps the fgs in the original unit cell to the
        equivalent fgs in the cell.

    """

    def __init__(self, id_, fgs):
        self.id = np.array(id_)
        self.fgs = fgs


class ChargedMolError(Exception):
    def __init__(self, mol_file, msg):
        self.mol_file = mol_file
        self.msg = msg


class MolFileError(Exception):
    def __init__(self, mol_file, msg):
        self.mol_file = mol_file
        self.msg = msg


class PopulationSizeError(Exception):
    def __init__(self, msg):
        self.msg = msg


class LazyAttr:
    """
    A descriptor for creating lazy attributes.

    """

    def __init__(self, func):
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            return self
        val = self.func(obj)
        setattr(obj, self.func.__name__, val)
        return val


class MAEExtractor:
    """
    Extracts the lowest energy conformer from a .maegz file.

    Macromodel conformer searches produce -out.maegz files containing
    all of the conformers found during the search and their energies
    and other data.

    Initializing this class with a :class:`.ConstructedMolecule` finds
    the ``-out.maegz`` file of that :class:`.ConstructedMolecule` and
    converts it to a ``.mae`` file. It then creates and additional
    ``.mae`` file holding only the lowest energy conformer found.

    Attributes
    ----------
    maegz_path : :class:`str`
        The path to the ``-out.maegz`` file generated by the macromodel
        conformer search.

    mae_path : :class:`str`
        The path to the ``.mae`` file holding the conformers generated
        by the macromodel conformer search.

    content : :class:`str`
        The content of the ``.mae`` file hodling all the conformers
        from the macromodel conformer search. This holds other data
        such as their energies too.

    energies : :class:`list`
        The :class:`list` has the form

        .. code-block:: python

            energies = [(0, 231.0), (1, 144.4), ...]

        Each :class:`tuple` holds the id and energy of every conformer
        in the ``.mae`` file, respectively.

    min_energy : :class:`float`
        The minimum energy found in the ``.mae`` file.

    path : :class:`str`
        The full path of the ``.mae`` file holding the extracted lowest
        energy conformer.

    """

    def __init__(self, run_name, n=1):
        self.maegz_path = f'{run_name}-out.maegz'
        self.maegz_to_mae()
        self.extract_conformers(n)

    def extract_conformers(self, n):
        """
        Creates ``.mae`` files holding the lowest energy conformers.

        Parameters
        ----------
        n : :class:`int`
            The number of conformers to extract.

        Returns
        -------
        None : :class:`NoneType`

        """

        for i in range(n):
            # Get the id of the lowest energy conformer.
            num = self.lowest_energy_conformers(n)[i][1]
            # Get the structure block corresponding to the lowest
            # energy conformer.
            content = self.content.split("f_m_ct")
            new_mae = "f_m_ct".join([content[0], content[num]])

            # Write the structure block in its own .mae file, named
            # after conformer extracted.
            if n == 1:
                # Write the structure block in its own .mae file, named
                # after conformer extracted.
                new_name = self.mae_path.replace(
                    '.mae',
                    f'EXTRACTED_{num}.mae'
                )
            else:
                new_name = self.mae_path.replace(
                    '.mae',
                    f'EXTRACTED_{num}_conf_{i}.mae'
                )

            with open(new_name, 'w') as mae_file:
                mae_file.write(new_mae)

            if i == 0:
                # Save the path of the newly created file.
                self.path = new_name

    def extract_energy(self, block):
        """
        Extracts the energy value from a ``.mae`` energy data block.

        Parameters
        ----------
        block : :class:`str`
            An ``.mae`` energy data block.

        Returns
        -------
        :class:`float`
            The energy value extracted from `block` or ``None`` if
            one is not found.

        """

        block = block.split(":::")
        for name, value in zip(block[0].split('\n'),
                               block[1].split('\n')):
            if 'r_mmod_Potential_Energy' in name:
                return float(value)

    def lowest_energy_conformers(self, n):
        """
        Returns the id and energy of the lowest energy conformers.

        Parameters
        ----------
        n : :class:`int`
            The number of lowest energy conformers to return.

        Returns
        -------
        :class:`list`
            A :class:`list` of the form

            .. code-block:: python

                returned = [(23, 123.3), (1, 143.89), (12, 150.6), ...]

            Where each :class:`tuple` holds the id and energy of the
            `n` lowest energy conformers, respectively.

        """

        # Open the .mae file holding all the conformers and load its
        # content.
        with open(self.mae_path, 'r') as mae_file:
            self.content = mae_file.read()
            # Split the content across curly braces. This divides the
            # various sections of the .mae file.
            content_split = re.split(r"[{}]", self.content)

        # Go through all the datablocks in the the .mae file. For each
        # energy block extract the energy and store it in the
        # `energies` list. Store the `index`  (conformer id) along with
        # each extracted energy.
        self.energies = []
        prev_block = deque([""], maxlen=1)
        index = 1
        for block in content_split:
            if ("f_m_ct" in prev_block[0] and
               "r_mmod_Potential_Energy" in block):
                energy = self.extract_energy(block)
                self.energies.append((energy, index))
                index += 1

            prev_block.append(block)

        # Selecting the lowest energy n conformers
        confs = sorted(self.energies)[:n]
        # Define the energy of the lowest energy conformer
        self.min_energy = confs[0][0]
        # Return a list with id and energy of the lowest energy
        # conformers.
        return confs

    def maegz_to_mae(self):
        """
        Converts the .maegz file to a .mae file.

        Returns
        -------
        None : :class:`NoneType`

        """

        self.mae_path = self.maegz_path.replace('.maegz', '.mae')
        with gzip.open(self.maegz_path, 'r') as maegz_file:
            with open(self.mae_path, 'wb') as mae_file:
                mae_file.write(maegz_file.read())


def archive_output():
    """
    Places the ``output`` folder into ``old_output``.

    This function assumes that the ``output`` folder is in the current
    working directory.

    Returns
    -------
    None : :class:`NoneType`

    """

    if 'output' not in os.listdir():
        return

    # Make the ``old_output`` folder if it does not exist already.
    if 'old_output' not in os.listdir():
        os.mkdir('old_output')

    # Find out with what number the ``output`` folder should be
    # labelled within ``old_output``.
    num = len(os.listdir('old_output'))
    new_dir = os.path.join('old_output', str(num))
    os.rename('output', new_dir)


def dedupe(iterable, key=None, seen=None):
    """
    Yields items from `iterable` barring duplicates.

    If `seen` is provided it contains elements which are not to be
    yielded at all.

    Parameters
    ----------
    iterable : :class:`iterable`
        An iterable of elements which are to be yielded, only once.

    key : :class:`callable`
        A function which gets applied to every member of `iterable`.
        The return of this function is checked for duplication rather
        than the member itself.

    seen : :class:`set`, optional
        Holds items which are not to be yielded.

    Yields
    ------
    :class:`object`
        Element in `iterable` which is not in `seen` and has not been
        yielded before.

    """

    if seen is None:
        seen = set()
    for x in iterable:
        val = key(x) if key is not None else x
        if val not in seen:
            seen.add(val)
            yield x


def flatten(iterable, excluded_types=None):
    """
    Transforms an nested iterable into a flat one.

    For example

    .. code-block:: python

        [[1,2,3], [[4], [5],[[6]], 7]

    becomes

    .. code-block:: python

        [1,2,3,4,5,6,7]

    If a type is found in `excluded_types` it will not be yielded from.
    For example if ``str`` is in `excluded_types`

    .. code-block:: python

        a = ["abcd", ["efgh"]]

    "abcd" and "efgh" are yielded if `a` is passed to `iterable`. If
    `str` was not in `excluded_types` then "a", "b", "c", "d", "e",
    "f", "g" and "h" would all be yielded individually.

    Parameters
    ----------
    iterable : :class:`iterable`
        The iterable which is to be flattened

    excluded_types : :class:`set`, optional
        Holds container types which are not be flattened.

    Yields
    ------
    :class:`object`
        A nested element of `iterable`.

    """

    if excluded_types is None:
        excluded_types = {str}

    for x in iterable:
        if hasattr(x, '__iter__') and type(x) not in excluded_types:
            yield from flatten(x, excluded_types)
        else:
            yield x


def kabsch(coords1, coords2):
    """
    Return a rotation matrix to minimize dstance between 2 coord sets.

    This is essentially an implementation of the Kabsch algorithm.
    Given two sets of coordinates, `coords1` and `coords2`, this
    function returns a rotation matrix. When the rotation matrix is
    applied to `coords1` the resulting coordinates have their rms
    distance to `coords2` minimized.

    Parameters
    ----------
    coords1 : :class:`numpy.ndarray`
        This array represents a matrix hodling coordinates which need
        to be rotated to minimize their rms distance to coordinates in
        `coords2`. The matrix is n x 3. Each row of the matrix holds
        the x, y and z coordinates of one point, respectively. Here
        ``n`` is the number of points.

    coords2 : :class:`numpy.ndarray`
        This array represents a matrix which holds the coordinates of
        points the distance to which should be minimized. The matrix is
        ``n x 3``. Each row of the matrix holds the x, y and z
        coordinates of one point, respectively. Here ``n`` is the
        number of points.

    Returns
    -------
    :class:`numpy.ndarray`
        A rotation matrix. This will be a ``3 x 3`` matrix.

    References
    ----------
    http://nghiaho.com/?page_id=671
    https://en.wikipedia.org/wiki/Kabsch_algorithm

    """

    h = np.dot(coords1, coords2.T)
    u, s, v = np.linalg.svd(h)

    if int(np.linalg.det(v)) < 0:
        v[:, 2] = -v[:, 2]

    return np.dot(v, u)


def kill_macromodel():
    """
    Kills any applications left open as a result running MacroModel.

    Applications that are typically left open are
    ``jserver-watcher.exe`` and ``jservergo.exe``.

    Returns
    -------
    None : :class:`NoneType`

    """

    if os.name == 'nt':
        # In Windows, use the ``Taskkill`` command to force a close on
        # the applications.
        sp.run(["Taskkill", "/IM", "jserver-watcher.exe", "/F"],
               stdout=sp.PIPE, stderr=sp.PIPE)
        sp.run(["Taskkill", "/IM", "jservergo.exe", "/F"],
               stdout=sp.PIPE, stderr=sp.PIPE)

    if os.name == 'posix':
        sp.run(["pkill", "jservergo"],
               stdout=sp.PIPE, stderr=sp.PIPE)
        sp.run(["pkill", "jserver-watcher"],
               stdout=sp.PIPE, stderr=sp.PIPE)


def matrix_centroid(matrix):
    """
    Returns the centroid of the coordinates held in `matrix`.

    Parameters
    ----------
    matrix : :class:`np.ndarray`
        A ``n x 3`` matrix. Each row holds the x, y and z
        coordinate of some point, respectively.

    Returns
    -------
    :class:`numpy.ndarray`
        A numpy array which holds the x, y and z
        coordinates of the centroid of the coordinates in `matrix`.

    """

    return np.sum(matrix, axis=0) / len(matrix)


def mol_from_mae_file(mae_path):
    """
    Creates a ``rdkit`` molecule from a ``.mae`` file.

    Parameters
    ----------
    mol2_file : :class:`str`
        The full path of the ``.mae`` file from which an rdkit molecule
        should be instantiated.

    Returns
    -------
    :class:`rdkit.Mol`
        An ``rdkit`` instance of the molecule held in `mae_file`.

    """

    mol = rdkit.EditableMol(rdkit.Mol())
    conf = rdkit.Conformer()

    with open(mae_path, 'r') as mae:
        content = re.split(r'[{}]', mae.read())

    prev_block = deque([''], maxlen=1)
    for block in content:
        if 'm_atom[' in prev_block[0]:
            atom_block = block
        if 'm_bond[' in prev_block[0]:
            bond_block = block
        prev_block.append(block)

    labels, data_block, *_ = atom_block.split(':::')
    labels = [label for label in labels.split('\n')
              if not label.isspace() and label != '']

    data_block = [a.split() for a in data_block.split('\n') if
                  not a.isspace() and a != '']

    for line in data_block:
        line = [word for word in line if word != '"']
        if len(labels) != len(line):
            raise RuntimeError(('Number of labels does'
                                ' not match number of columns'
                                ' in .mae file.'))

        for label, data in zip(labels, line):
            if 'x_coord' in label:
                x = float(data)
            if 'y_coord' in label:
                y = float(data)
            if 'z_coord' in label:
                z = float(data)
            if 'atomic_number' in label:
                atom_num = int(data)

        atom_sym = periodic_table[atom_num]
        atom_coord = Point3D(x, y, z)
        atom_id = mol.AddAtom(rdkit.Atom(atom_sym))
        conf.SetAtomPosition(atom_id, atom_coord)

    labels, data_block, *_ = bond_block.split(':::')
    labels = [label for label in labels.split('\n')
              if not label.isspace() and label != '']
    data_block = [a.split() for a in data_block.split('\n')
                  if not a.isspace() and a != '']

    for line in data_block:
        if len(labels) != len(line):
            raise RuntimeError(('Number of labels does'
                                ' not match number of '
                                'columns in .mae file.'))

        for label, data in zip(labels, line):
            if 'from' in label:
                atom1 = int(data) - 1
            if 'to' in label:
                atom2 = int(data) - 1
            if 'order' in label:
                bond_order = str(int(data))
        mol.AddBond(atom1, atom2, bond_dict[bond_order])

    mol = mol.GetMol()
    mol.AddConformer(conf)
    return mol


def mol_from_mol_file(mol_file):
    """
    Creates a rdkit molecule from a ``.mol`` (V3000) file.

    Parameters
    ----------
    mol_file : :class:`str`
        The full of the .mol file from which an rdkit molecule should
        be instantiated.

    Returns
    -------
    :class:`rdkit.Mol`
        An rdkit instance of the molecule held in `mol2_file`.

    Raises
    ------
    :class:`ChargedMolError`
        If an atom row has more than 8 coloumns it is usually because
        there is a 9th coloumn indicating atomic charge. Such molecules
        are not currently supported, so an error is raised.

    :class:`MolFileError`
        If the file is not a V3000 ``.mol`` file.

    """

    e_mol = rdkit.EditableMol(rdkit.Mol())
    conf = rdkit.Conformer()

    with open(mol_file, 'r') as f:
        take_atom = False
        take_bond = False
        v3000 = False

        for line in f:
            if 'V3000' in line:
                v3000 = True

            if 'M  V30 BEGIN ATOM' in line:
                take_atom = True
                continue

            if 'M  V30 END ATOM' in line:
                take_atom = False
                continue

            if 'M  V30 BEGIN BOND' in line:
                take_bond = True
                continue

            if 'M  V30 END BOND' in line:
                take_bond = False
                continue

            if take_atom:
                words = line.split()
                if len(words) > 8:
                    raise ChargedMolError(mol_file,
                                          ('Atom row has more'
                                           ' than 8 coloumns. Likely '
                                           'due to a charged atom.'))
                _, _, _, atom_sym, *coords, _ = words
                coords = [float(x) for x in coords]
                atom_coord = Point3D(*coords)
                atom_id = e_mol.AddAtom(rdkit.Atom(atom_sym))
                conf.SetAtomPosition(atom_id, atom_coord)
                continue

            if take_bond:
                *_, bond_id, bond_order, atom1, atom2 = line.split()
                e_mol.AddBond(int(atom1)-1, int(atom2)-1,
                              bond_dict[bond_order])
                continue
    if not v3000:
        raise MolFileError(mol_file, 'Not a V3000 .mol file.')

    mol = e_mol.GetMol()
    mol.AddConformer(conf)
    return mol


def move_generated_macromodel_files(basename, output_dir):
    if not os.path.exists(output_dir):
        os.mkdir(output_dir)

    for filename in iglob(f'{basename}*'):
        # Do not move the output_dir.
        if filename == output_dir:
            continue
        os.rename(filename, f'{output_dir}/{filename}')


def normalize_vector(vector):
    """
    Normalizes the given vector.

    A new vector is returned, the original vector is not modified.

    Parameters
    ----------
    vector : :class:`np.ndarray`
        The vector to be normalized.

    Returns
    -------
    :class:`np.ndarray`
        The normalized vector.

    """

    return np.divide(vector, np.linalg.norm(vector))


def remake(mol):
    """
    Remakes a molecule from scratch.

    Parameters
    ----------
    mol : :class:`rdkit.Mol`
        The molecule to be remade.

    Returns
    -------
    :class:`rdkit.Mol`
        The remade molecule.

    """

    emol = rdkit.EditableMol(rdkit.Mol())
    for atom in mol.GetAtoms():
        new = rdkit.Atom(atom.GetAtomicNum())
        new.SetFormalCharge(atom.GetFormalCharge())
        emol.AddAtom(new)

    for bond in mol.GetBonds():
        emol.AddBond(
            beginAtomIdx=bond.GetBeginAtomIdx(),
            endAtomIdx=bond.GetEndAtomIdx(),
            order=bond.GetBondType()
        )

    m = emol.GetMol()
    m.AddConformer(mol.GetConformer())
    return m


def orthogonal_vector(vector):
    ortho = [0, 0, 0]
    for m, val in enumerate(vector):
        if not np.allclose(val, 0, atol=1e-8):
            n = (m+1) % 3
            break
    ortho[n] = vector[m]
    ortho[m] = -vector[n]
    return ortho


def rotation_matrix(vector1, vector2):
    """
    Returns a rotation matrix which transforms `vector1` to `vector2`.

    Multiplying `vector1` by the rotation matrix returned by this
    function yields `vector2`.

    Parameters
    ----------
    vector1 : :class:`numpy.ndarray`
        The vector which needs to be transformed to `vector2`.

    vector2 : :class:`numpy.ndarray`
        The vector onto which `vector1` needs to be transformed.

    Returns
    -------
    :class:`numpy.ndarray`
        A rotation matrix which transforms `vector1` to `vector2`.

    References
    ----------
    http://tinyurl.com/kybj9ox
    http://tinyurl.com/gn6e8mz

    """

    # Make sure both inputs are unit vectors.
    vector1 = normalize_vector(vector1)
    vector2 = normalize_vector(vector2)

    # Hande the case where the input and output vectors are equal.
    if np.allclose(vector1, vector2, atol=1e-8):
        return np.identity(3)

    # Handle the case where the rotation is 180 degrees.
    if np.allclose(vector1, np.multiply(vector2, -1), atol=1e-8):
        return rotation_matrix_arbitrary_axis(
            angle=np.pi,
            axis=orthogonal_vector(vector1)
        )

    v = np.cross(vector1, vector2)
    vx = np.array([
        [0, -v[2], v[1]],
        [v[2], 0, -v[0]],
        [-v[1], v[0], 0]
    ])
    s = np.linalg.norm(v)
    c = np.dot(vector1, vector2)
    i = np.identity(3)
    mult_factor = (1-c)/np.square(s)
    return i + vx + np.multiply(np.dot(vx, vx), mult_factor)


def rotation_matrix_arbitrary_axis(angle, axis):
    """
    Returns a rotation matrix of `angle` radians about `axis`.

    Parameters
    ----------
    angle : :class:`float`
        The size of the rotation in radians.

    axis : :class:`numpy.ndarray`
        A 3 element aray which represents a vector. The vector is the
        axis about which the rotation is carried out.

    Returns
    -------
    :class:`numpy.ndarray`
        A ``3x3`` array representing a rotation matrix.

    """

    axis = normalize_vector(axis)

    a = np.cos(angle/2)
    b, c, d = axis * np.sin(angle/2)

    e11 = np.square(a) + np.square(b) - np.square(c) - np.square(d)
    e12 = 2*(b*c - a*d)
    e13 = 2*(b*d + a*c)

    e21 = 2*(b*c + a*d)
    e22 = np.square(a) + np.square(c) - np.square(b) - np.square(d)
    e23 = 2*(c*d - a*b)

    e31 = 2*(b*d - a*c)
    e32 = 2*(c*d + a*b)
    e33 = np.square(a) + np.square(d) - np.square(b) - np.square(c)

    return np.array([[e11, e12, e13],
                     [e21, e22, e23],
                     [e31, e32, e33]])


def dice_similarity(mol1, mol2, fp_radius=3):
    """
    Return the chemical similarity between two molecules.

    Parameters
    ----------
    mol1 : :class:`.Molecule`
        The first molecule.

    mol2 : :class:`.Molecule`
        The second molecule.

    fp_radius : :class:`int`, optional
        The radius of the Morgan fingerprint used to calculate
        similarity.

    Returns
    -------
    :class:`float`
        The similarity.

    """

    rdkit_mol1 = mol1.to_rdkit_mol()
    rdkit.SanitizeMol(rdkit_mol1)
    fp1 = rdkit.GetMorganFingerprint(
        mol=rdkit_mol1,
        radius=fp_radius,

    )
    rdkit_mol2 = mol2.to_rdkit_mol()
    rdkit.SanitizeMol(rdkit_mol2)
    fp2 = rdkit.GetMorganFingerprint(
        mol=rdkit_mol2,
        radius=fp_radius,

    )
    return rdkit.DataStructs.DiceSimilarity(fp1, fp2)


def quaternion(u):
    """
    Returns a translation + rotation quaternion.

    Parameters
    ----------
    u : :class:`list` of :class:`float`
        A :class:`list` of length 3 holding the parameter for the
        quarternion.

    References
    ----------
    K. Shoemake, Uniform random rotations, Graphics Gems III,
    pages 124-132. Academic, New York, 1992.

    """

    a, b, c = u
    q = np.zeros(4, np.float64)
    q[0] = np.sqrt(1. - a) * np.sin(2. * np.pi * b)
    q[1] = np.sqrt(1. - a) * np.cos(2. * np.pi * b)
    q[2] = np.sqrt(a) * np.sin(2. * np.pi * c)
    q[3] = np.sqrt(a) * np.cos(2. * np.pi * c)
    return q


def translation_component(q):
    """
    Extracts translation vector from quaternion.

    Parameters
    ----------
    q : :class:`numpy.ndarray`
        A length 4 quaternion.

    Returns
    -------
    :class:`numpy.ndarray`
        The translation vector encoded within `q`.

    """

    rot_epsilon = 1e-6
    q = np.copy(q)
    if q[0] < 0.:
        q = -q
    if q[0] > 1.0:
        q /= np.sqrt(np.dot(q, q))

    theta = 2. * np.arccos(q[0])
    s = np.sqrt(1. - q[0] * q[0])
    if s < rot_epsilon:
        p = 2. * q[1:4]
    else:
        p = q[1:4] / s * theta
    return p


def tar_output():
    """
    Places all the content in the ``output`` folder into a .tgz file.

    Returns
    -------
    None : :class:`NoneType`

    """

    tname = os.path.join('output', 'output.tgz')
    with tarfile.open(tname, 'w:gz') as tar:
        tar.add('output')


@contextmanager
def time_it():
    """
    Times the code executed within the indent.

    This is a context manager so it should be used as:

    .. code-block:: python

        with time_it():
            something1()
            something2()
            something3()

    After all 3 functions are finished and the nested block is exited
    the time taken to process the entire block is printed.

    """

    start = time.time()
    yield
    time_taken = time.time() - start
    m, s = divmod(time_taken, 60)
    h, m = divmod(m, 60)
    print(f'\nTime taken was {int(h)} : {int(m)} : {int(s)}.\n\n')


def vector_angle(vector1, vector2):
    """
    Returns the angle between two vectors in radians.

    Parameters
    ----------
    vector1 : :class:`numpy.ndarray`
        The first vector.

    vector2 : :class:`numpy.ndarray`
        The second vector.

    Returns
    -------
    :class:`float`
        The angle between `vector1` and `vector2` in radians.

    """

    numerator = np.dot(vector1, vector2)
    denominator = (np.linalg.norm(vector1) * np.linalg.norm(vector2))
    # This if statement prevents returns of NaN due to floating point
    # incurracy.
    term = numerator/denominator
    if np.isclose(term, 1, atol=1e-12):
        return 0.0
    if np.isclose(term, -1, atol=1e-12):
        return np.pi
    return np.arccos(term)


class XTBInvalidSolventError(Exception):
    ...


def is_valid_xtb_solvent(gfn_version, solvent):
    """
    Check if solvent is valid for the given GFN version.

    Parameters
    ----------
    gfn_version : :class:`int`
        GFN parameterization version. Can be: ``0``, ``1`` or ``2``.

    solvent : :class:`str`
        Solvent being tested [1]_.

    Returns
    -------
    :class:`bool`
        ``True`` if solvent is valid.

    References
    ----------
    .. [1] https://xtb-docs.readthedocs.io/en/latest/gbsa.html

    """
    if gfn_version == 0:
        return False
    elif gfn_version == 1:
        valid_solvents = {
            'acetone', 'acetonitrile', 'benzene',
            'CH2Cl2'.lower(), 'CHCl3'.lower(), 'CS2'.lower(),
            'DMSO'.lower(), 'ether', 'H2O'.lower(),
            'methanol', 'THF'.lower(), 'toluene'
        }
        return solvent in valid_solvents
    elif gfn_version == 2:
        valid_solvents = {
            'acetone', 'acetonitrile', 'CH2Cl2'.lower(),
            'CHCl3'.lower(), 'CS2'.lower(), 'DMF'.lower(),
            'DMSO'.lower(), 'ether', 'H2O'.lower(), 'methanol',
            'n-hexane'.lower(), 'THF'.lower(), 'toluene'
        }
        return solvent in valid_solvents


class XTBExtractor:
    """
    Extracts properties from GFN-xTB output files.

    Attributes
    ----------
    output_file : :class:`str`
        Output file to extract properties from.

    output_lines : :class:`list` : :class:`str`
        :class:`list` of all lines in as :class:`str` in the output
        file.

    total_energy : :class:`float`
        The total energy in the :attr:`output_file` as
        :class:`float`. The energy is in units of a.u..

    homo_lumo_gap : :class:`float`
        The HOMO-LUMO gap in the :attr:`output_file` as
        :class:`float`. The gap is in units of eV.

    fermi_level : :class:`float`
        The Fermi level in the :attr:`output_file` as
        :class:`float` in units of eV.

    qonly_dipole_moment : :class:`list`
        Components of the Q only dipole moment in units
        of Debye in :class:`list` of the form
        ``[x, y, z]``.

    full_dipole_moment : :class:`list`
        Components of the full dipole moment in units
        of Debye in :class:`list` of the form
        ``[x, y, z, total]``.

    qonly_quadrupole_moment : :class:`list`
        Components of the Q only traceless quadrupole moment in units
        of Debye in :class:`list` of the form
        ``[xx, xy, xy, xz, yz, zz]``.

    qdip_quadrupole_moment : :class:`list`
        Components of the Q+Dip traceless quadrupole moment in units of
        Debye in :class:`list` of the form
        ``[xx, xy, xy, xz, yz, zz]``.

    full_quadrupole_moment : :class:`list`
        Components of the full traceless quadrupole moment in units of
        Debye in :class:`list` of the form
        ``[xx, xy, xy, xz, yz, zz]``.

    homo_lumo_occ : :class:`dict`
        :class:`dict` of :class:`list` containing the orbital number,
        energy in eV and occupation of the HOMO and LUMO orbitals in
        the :attr:`output_file`.

    total_free_energy : :class:`float`
        The total free energy in the :attr:`output_file` as
        :class:`float`. The free energy is in units of a.u. and
        calculated at 298.15K.

    frequencies : :class:`list`
        :class:`list` of the vibrational frequencies as :class:`float`
        in the :attr:`output_file`. Vibrational frequencies are in
        units of wavenumber and calculated at 298.15K.

    """
    def __init__(self, output_file):
        """
        Initializes :class:`XTBExtractor`

        Parameters
        ----------
        output_file : :class:`str`
            Output file to extract properties from.

        """
        self.output_file = output_file
        # Explictly set encoding to UTF-8 because default encoding on
        # Windows will fail to read the file otherwise.
        with open(self.output_file, 'r', encoding='UTF-8') as f:
            self.output_lines = f.readlines()

        self._extract_values()

    def _extract_values(self):
        """
        Extract all properties from xTB output file.

        Returns
        -------
        None : :class:`NoneType`

        """

        for i, line in enumerate(self.output_lines):
            if self._check_line(line, 'total_energy'):
                self._extract_total_energy(line)
            elif self._check_line(line, 'homo_lumo_gap'):
                self._extract_homo_lumo_gap(line)
            elif self._check_line(line, 'fermi_level'):
                self._extract_fermi_level(line)
            elif self._check_line(line, 'dipole_moment'):
                self._extract_qonly_dipole_moment(i)
                self._extract_full_dipole_moment(i)
            elif self._check_line(line, 'quadrupole_moment'):
                self._extract_qonly_quadrupole_moment(i)
                self._extract_qdip_quadrupole_moment(i)
                self._extract_full_quadrupole_moment(i)
            elif self._check_line(line, 'homo_lumo_occ_HOMO'):
                self.homo_lumo_occ = {}
                self._extract_homo_lumo_occ(line, 'HOMO')
            elif self._check_line(line, 'homo_lumo_occ_LUMO'):
                self._extract_homo_lumo_occ(line, 'LUMO')
            elif self._check_line(line, 'total_free_energy'):
                self._extract_total_free_energy(line)

        # Frequency formatting requires loop through full file.
        self._extract_frequencies()

    def _check_line(self, line, option):
        """
        Checks a line for a string based on option.

        All formatting based on the 190418 version of xTB.

        Parameters
        ----------
        line : :class:`str`
            Line of output file to check.

        option : :class:`str`
            Define which property and string being checked for.
            Can be one of ``'total_energy'``, ``'homo_lumo_gap'``,
            ``'fermi_level'``, ``'dipole_moment'``,
            ``'quadrupole_moment'``, ``'homo_lumo_occ_HOMO'``,
            ``'homo_lumo_occ_LUMO'``,
            ``'total_free_energy'``.

        Returns
        -------
        :class:`bool`
            Returns ``True`` if the desired string is present.

        """
        options = {
            'total_energy': '          | TOTAL ENERGY  ',
            'homo_lumo_gap': '          | HOMO-LUMO GAP   ',
            'fermi_level': '             Fermi-level        ',
            'dipole_moment': 'molecular dipole:',
            'quadrupole_moment': 'molecular quadrupole (traceless):',
            'homo_lumo_occ_HOMO': '(HOMO)',
            'homo_lumo_occ_LUMO': '(LUMO)',
            'total_free_energy': '          | TOTAL FREE ENERGY  ',
        }

        if options[option] in line:
            return True

    def _extract_total_energy(self, line):
        """
        Updates :attr:`total_energy`.

        Parameters
        ----------
        line : :class:`str`
            Line of output file to extract property from.

        Returns
        -------
        None : :class:`NoneType`

        """

        nums = re.compile(r"[+-]?\d+(?:\.\d+)?(?:[eE][+-]?\d+)?")
        string = nums.search(line.rstrip()).group(0)
        self.total_energy = float(string)

    def _extract_homo_lumo_gap(self, line):
        """
        Updates :attr:`homo_lumo_gap`.

        Parameters
        ----------
        line : :class:`str`
            Line of output file to extract property from.

        Returns
        -------
        None : :class:`NoneType`

        """

        nums = re.compile(r"[+-]?\d+(?:\.\d+)?(?:[eE][+-]?\d+)?")
        string = nums.search(line.rstrip()).group(0)
        self.homo_lumo_gap = float(string)

    def _extract_fermi_level(self, line):
        """
        Updates :attr:`fermi_level`.

        Parameters
        ----------
        line : :class:`str`
            Line of output file to extract property from.

        Returns
        -------
        None : :class:`NoneType`

        """

        nums = re.compile(r"[+-]?\d+(?:\.\d+)?(?:[eE][+-]?\d+)?")
        part2 = line.split('Eh')
        string = nums.search(part2[1].rstrip()).group(0)
        self.fermi_level = float(string)

    def _extract_qonly_dipole_moment(self, index):
        """
        Updates :attr:`qonly_dipole_moment`.

        Parameters
        ----------
        line : :class:`str`
            Line of output file to extract property from.

        index : :class:`int`
            Index of line in :attr:`output_lines`.

        Returns
        -------
        None : :class:`NoneType`

        """

        sample_set = self.output_lines[index+2].rstrip()

        if 'q only:' in sample_set:
            self.qonly_dipole_moment = [
                float(i)
                for i in sample_set.split(':')[1].split(' ') if i
            ]

    def _extract_full_dipole_moment(self, index):
        """
        Updates :attr:`full_dipole_moment`.

        Parameters
        ----------
        line : :class:`str`
            Line of output file to extract property from.

        index : :class:`int`
            Index of line in :attr:`output_lines`.

        Returns
        -------
        None : :class:`NoneType`

        """

        sample_set = self.output_lines[index+3].rstrip()

        if 'full:' in sample_set:
            self.full_dipole_moment = [
                float(i)
                for i in sample_set.split(':')[1].split(' ') if i
            ]

    def _extract_qonly_quadrupole_moment(self, index):
        """
        Updates :attr:`qonly_quadrupole_moment`.

        Parameters
        ----------
        line : :class:`str`
            Line of output file to extract property from.

        index : :class:`int`
            Index of line in :attr:`output_lines`.

        Returns
        -------
        None : :class:`NoneType`

        """

        sample_set = self.output_lines[index+2].rstrip()

        if 'q only:' in sample_set:
            self.qonly_quadrupole_moment = [
                float(i)
                for i in sample_set.split(':')[1].split(' ') if i
            ]

    def _extract_qdip_quadrupole_moment(self, index):
        """
        Updates :attr:`qdip_quadrupole_moment`.

        Parameters
        ----------
        line : :class:`str`
            Line of output file to extract property from.

        index : :class:`int`
            Index of line in :attr:`output_lines`.

        Returns
        -------
        None : :class:`NoneType`

        """

        sample_set = self.output_lines[index+3].rstrip()

        if 'q+dip:' in sample_set:
            self.qdip_quadrupole_moment = [
                float(i)
                for i in sample_set.split(':')[1].split(' ') if i
            ]

    def _extract_full_quadrupole_moment(self, index):
        """
        Updates :attr:`full_quadrupole_moment`.

        Parameters
        ----------
        line : :class:`str`
            Line of output file to extract property from.

        index : :class:`int`
            Index of line in :attr:`output_lines`.

        Returns
        -------
        None : :class:`NoneType`

        """

        sample_set = self.output_lines[index+4].rstrip()

        if 'full:' in sample_set:
            self.full_quadrupole_moment = [
                float(i)
                for i in sample_set.split(':')[1].split(' ') if i
            ]

    def _extract_homo_lumo_occ(self, line, orbital):
        """
        Updates :attr:`homo_lumo_occ`.

        Parameters
        ----------
        line : :class:`str`
            Line of output file to extract property from.

        orbital : :class:`str`
            Can be 'HOMO' or 'LUMO'.

        Returns
        -------
        None : :class:`NoneType`

        """

        if orbital == 'HOMO':
            split_line = [i for i in line.rstrip().split(' ') if i]
            # The line is:
            #   Number, occupation, energy (Ha), energy (ev), label
            # Extract:
            #   Number, occupation, energy (eV)
            orbital_val = [
                int(split_line[0]),
                float(split_line[1]),
                float(split_line[3])
            ]
        elif orbital == 'LUMO':
            split_line = [i for i in line.rstrip().split(' ') if i]
            # The line is:
            #   Number, energy (Ha), energy (ev), label
            # Extract:
            #   Number, occupation (zero), energy (eV)
            orbital_val = [
                int(split_line[0]),
                0,
                float(split_line[2])
            ]

        self.homo_lumo_occ[orbital] = orbital_val

    def _extract_total_free_energy(self, line):
        """
        Updates :attr:`total_free_energy`.

        Returns
        -------
        None : :class:`NoneType`

        """

        nums = re.compile(r"[+-]?\d+(?:\.\d+)?(?:[eE][+-]?\d+)?")
        string = nums.search(line.rstrip()).group(0)
        self.total_free_energy = float(string)

    def _extract_frequencies(self):
        """
        Updates :attr:`frequencies`.

        Returns
        -------
        None : :class:`NoneType`

        """

        test = '|               Frequency Printout                |'

        # Use a switch to make sure we are extracting values after the
        # final property readout.
        switch = False

        frequencies = []
        for i, line in enumerate(self.output_lines):
            if test in line:
                # Turn on reading as final frequency printout has
                # begun.
                switch = True
            if ' reduced masses (amu)' in line:
                # Turn off reading as frequency section is done.
                switch = False
            if 'eigval :' in line and switch is True:
                samp = line.rstrip().split(':')[1].split(' ')
                split_line = [i for i in samp if i]
                for freq in split_line:
                    frequencies.append(freq)

        self.frequencies = [float(i) for i in frequencies]



from .topology_graphs import (
    cof,
    macrocycle,
    polymer,
    rotaxane,
    cage,
    host_guest
)
from .topology_graphs.topology_graph import *
from .functional_groups import *
from .molecules import *
from .elements import *
from .bonds import *



"""
Defines :class:`.Reactor`.

.. _`adding complex reactions`:

Extending stk: Adding complex reactions.
----------------------------------------

See :class:`.Reactor`.

"""

from collections import Counter
import numpy as np
from scipy.spatial.distance import euclidean

from . import elements
from .bonds import Bond


class _ReactionKey:
    """
    Represents a hash value of a collection of functional group names.

    The functional group names indicate which functional groups are
    involved in a reaction. The order in which the names are provided
    does not change the hash value. The quantity of each functional
    group name does change the hash value. Each name can be provided
    multiple times. Adding the same name again will change the hash
    value.

    Examples
    --------
    The key is constant with respect to permutations

    .. code-block:: python

        key1 = _ReactionKey('amine', 'aldehyde', 'amine')
        key2 = _ReactionKey('amine', 'amine', 'aldehyde')
        key3 = _ReactionKey('aldehyde', 'amine', 'amine')

        key1 == key2  # This is True.
        key1 == key3  # This is True.
        key2 == key3  # This is True.

    The key changes if the number of functional groups changes

    .. code-block:: python

        key4 = _ReactionKey('amine', 'aldehyde')
        key4 != key1  # This is True.


    """

    __slots__ = ['_key']

    def __init__(self, *fg_names):
        """
        Initialize a :class:`_ReactionKey`.

        Parameters
        ----------
        *fg_names : :class:`str`
            The names of functional groups involved in a reaction. If
            multiples of a functional group are present in a reaction,
            the name of the functional group must be present multiple
            times.

        """

        c = Counter(fg_names)
        # _key is the hash value for the reaction. It should change
        # if the different numbers of functional groups are present
        # in the reaction, but it should not change if the order in
        # which the functional groups are given changes.
        self._key = tuple(
            sorted((key, value) for key, value in c.items())
        )

    def __eq__(self, other):
        return self._key == other._key

    def __hash__(self):
        return hash(self._key)

    def __repr__(self):
        fg_names = ', '.join(
            repr(name)
            for name, count in self._key
            for i in range(count)
        )
        return f'_ReactionKey({fg_names})'

    def __str__(self):
        return repr(self)


class Reactor:
    """
    Performs reactions between functional groups of a molecule.

    The :class:`Reactor` is initialized with the
    :class:`.ConstructedMolecule` instance on which it will perform
    reactions.

    .. code-block:: python

        reactor = Reactor(mol)

    The :class:`Reactor` registers which functional groups need to have
    reactions performed with :meth:`add_reaction`

    .. code-block:: python

        fg1, fg2, fg3, fg4, fg5, fg6, fg7 = mol.func_groups

        # Register a reaction between the atoms in fg1 and fg2.
        reactor.add_reaction(fg1, fg2)

        # Register a reaction between atoms in fg3 and fg4.
        reactor.add_reaction(fg3, fg4)

        # Some reactions can take multiple functional groups.
        # You can put in as many functional groups as you like, given
        # an appropriate reaction is defined.
        reactor.add_reaction(fg5, fg6, fg7)

    Once all the reactions have been registered, they are perfomed
    in a single step

    .. code-block:: python

        reactor.finalize()

    An obvious question given this tutorial, is what reaction does
    :meth:`react` carry out? This is documented by :meth:`react`.
    However, :meth:`react` in most cases, will carry out a default
    reaction, which adds a bond between the bonder atoms of two
    functional groups. The bond order of the added bond is single by
    default but can be modified by editing :attr:`_bond_orders`. Here
    you will specify the :class:`_ReactionKey` for a reaction and what
    bond order you want that reaction to use.

    For some reactions you may wish to forgo the default reaction and
    do something more complex. This is neccessary because some
    reactions cannot be described by the simple combination of adding a
    bond while deleting some existing atoms. For example,
    consider the aldol reaction:

        CH3C(=O)CH3 + CH3C(=O)CH3 --> CH3(=O)CH2C(OH)(CH3)CH3

    Here a ketone is converted into an alcohol. If you wish to
    support a complex reaction, add it as a method within this class.
    The method will need take some :class:`FunctionalGroup` instances
    as arguments. These are the functional groups which react.

    Once the method is defined, :attr:`_custom_reactions` needs to
    be updated.

    """

    # When the default reaction is added by add_reaction(),
    # if the bond added between the two functional groups is not
    # single, the desired bond order should be placed in this
    # dictionary. The dictionary maps the reaction's
    # _ReactionKey to the desired bond order.
    _bond_orders = {
        _ReactionKey('amine', 'aldehyde'): 2,
        _ReactionKey('amide', 'aldehyde'): 2,
        _ReactionKey('nitrile', 'aldehyde'): 2,
        _ReactionKey('amide', 'amine'): 2,
        _ReactionKey('terminal_alkene', 'terminal_alkene'): 2,
        _ReactionKey('alkyne2', 'alkyne2'): 3
    }

    def __init__(self, mol):
        """
        Initialize a :class:`Reactor`.

        Parameters
        ----------
        mol : :class:`.ConstructedMolecule`
            The molecule on which the reactor adds and removes atoms
            and bonds.

        """

        # The molecule from which atoms and bonds are added and
        # removed.
        self._mol = mol
        # The ids of atoms which are to be removed.
        self._deleter_ids = set()
        # The ids of bonds which are to be removed.
        self._deleter_bonds = set()

        # Maps a _ReactionKey for a given reaction to a custom
        # method which carries out the reaction. This means that
        # add_reaction will use that method for carrying out that
        # reaction instead.
        self._custom_reactions = {

            _ReactionKey('boronic_acid', 'diol'):
                self._react_boronic_acid_with_diol,

            _ReactionKey('diol', 'difluorene'):
                self._react_diol_with_dihalogen,

            _ReactionKey('diol', 'dibromine'):
                self._react_diol_with_dihalogen,

            _ReactionKey('ring_amine', 'ring_amine'):
                self._react_ring_amine_with_ring_amine

        }

    def add_reaction(self, func_groups, periodicity):
        """
        Create bonds between functional groups.

        This function first looks at the functional groups provided via
        the `func_groups` argument and checks which functional groups
        are involved in the reaction. If the functional groups are
        handled by one of the custom reactions specified in
        :attr:`_custom_reactions` then that function is executed.

        In all other cases the function is assumed to have received two
        functional groups to react. In these functional groups, the
        bonder atoms have a bond added. The bond is single,
        unless otherwise specified in :attr:`_bond_orders`.

        Parameters
        ----------
        func_groups : :class:`tuple` of :class:`.FunctionalGroup`
            The functional groups to react.

        periodicity : :class:`tuple` of :class:`int`
            Specifies the periodicity of the bonds added by the
            reaction, which bridge the `func_groups`. See
            :attr:`.Bond.periodicity`.

        Returns
        -------
        None : :class:`NoneType`

        """

        names = (fg.fg_type.name for fg in func_groups)
        reaction_key = _ReactionKey(*names)
        reaction = self._custom_reactions.get(
            reaction_key,
            self._react_any
        )
        return reaction(reaction_key, func_groups, periodicity)

    def finalize(self):
        """
        Finish performing reactions.

        Returns
        -------
        None : :class:`NoneType`

        """

        self._mol.atoms = [
            atom for atom in self._mol.atoms
            if atom.id not in self._deleter_ids
        ]
        self._mol.bonds = [
            bond for bond in self._mol.bonds
            if bond.atom1.id not in self._deleter_ids
            and bond.atom2.id not in self._deleter_ids
        ]
        self._mol._position_matrix = [
            row for i, row in enumerate(self._mol._position_matrix)
            if i not in self._deleter_ids
        ]

    def _remove_deleters(self, func_groups):
        """
        Remove deleter atoms from `fgs`.

        Parameters
        ----------
        func_groups : :class:`list` of :class:`.FunctionalGroup`
            The functional groups from which deleter atoms should be
            removed.

        Returns
        -------
        None : :class:`NoneType`

        """

        for fg in func_groups:
            self._deleter_ids.update(fg.get_deleter_ids())
            fg.atoms = tuple(
                a for a in fg.atoms if a.id not in self._deleter_ids
            )
            fg.deleters = ()

    def _react_any(self, reaction_key, func_groups, periodicity):
        """
        Create bonds between functional groups.

        Parameters
        ----------
        reaction_key : :class:`._ReactionKey`
            The key for the reaction.

        func_groups : :class:`list` of :class:`.FunctionalGroup`
            The functional groups from which deleter atoms should be
            removed.

        periodicity : :class:`tuple` of :class:`int`
            Specifies the periodicity of the bonds added by the
            reaction, which bridge the `func_groups`. See
            :attr:`.Bond.periodicity`.

        Returns
        -------
        None : :class:`NoneType`

        """

        self._remove_deleters(func_groups)

        fg1, fg2 = func_groups
        bond_order = self._bond_orders.get(reaction_key, 1)
        bond = Bond(
            atom1=fg1.bonders[0],
            atom2=fg2.bonders[0],
            order=bond_order,
            periodicity=periodicity
        )
        self._mol.bonds.append(bond)
        self._mol.construction_bonds.append(bond)

    def _react_diol_with_dihalogen(
        self,
        reaction_key,
        func_groups,
        periodicity
    ):
        """
        Create bonds between functional groups.

        Parameters
        ----------
        reaction_key : :class:`._ReactionKey`
            The key for the reaction.

        func_groups : :class:`list` of :class:`.FunctionalGroup`
            The functional groups from which deleter atoms should be
            removed.

        periodicity : :class:`tuple` of :class:`int`
            Specifies the periodicity of the bonds added by the
            reaction, which bridge the `func_groups`. See
            :attr:`.Bond.periodicity`.

        Returns
        -------
        None : :class:`NoneType`

        """

        self._remove_deleters(func_groups)

        fg1, fg2 = func_groups
        diol = fg1 if fg1.fg_type.name == 'diol' else fg2
        dihalogen = fg2 if diol is fg1 else fg1

        distances = []
        for carbon in dihalogen.get_bonder_ids():
            c_coord = self._mol._position_matrix[carbon]
            for oxygen in diol.get_bonder_ids():
                o_coord = self._mol._position_matrix[oxygen]
                d = euclidean(c_coord, o_coord)
                distances.append((d, carbon, oxygen))
        distances.sort()

        deduped_pairs = []
        seen_o, seen_c = set(), set()
        for d, c, o in distances:
            if c not in seen_c and o not in seen_o:
                deduped_pairs.append((c, o))
                seen_c.add(c)
                seen_o.add(o)

        (c1, o1), (c2, o2), *_ = deduped_pairs
        assert c1 != c2 and o1 != o2

        bond1 = Bond(
            atom1=self._mol.atoms[c1],
            atom2=self._mol.atoms[o1],
            order=1,
            periodicity=periodicity
        )
        bond2 = Bond(
            atom1=self._mol.atoms[c2],
            atom2=self._mol.atoms[o2],
            order=1,
            periodicity=periodicity
        )
        self._mol.bonds.append(bond1)
        self._mol.construction_bonds.append(bond1)
        self._mol.bonds.append(bond2)
        self._mol.construction_bonds.append(bond2)

    def _react_boronic_acid_with_diol(
        self,
        reaction_key,
        func_groups,
        periodicity
    ):
        """
        Create bonds between functional groups.

        Parameters
        ----------
        reaction_key : :class:`._ReactionKey`
            The key for the reaction.

        func_groups : :class:`list` of :class:`.FunctionalGroup`
            The functional groups from which deleter atoms should be
            removed.

        periodicity : :class:`tuple` of :class:`int`
            Specifies the periodicity of the bonds added by the
            reaction, which bridge the `func_groups`. See
            :attr:`.Bond.periodicity`.

        Returns
        -------
        None : :class:`NoneType`

        """

        self._remove_deleters(func_groups)
        fg1, fg2 = func_groups
        boron = fg1 if fg1.fg_type.name == 'boronic_acid' else fg2
        diol = fg2 if boron is fg1 else fg1

        boron_atom = boron.bonders[0]
        bond1 = Bond(boron_atom, diol.bonders[0], 1, periodicity)
        self._mol.bonds.append(bond1)
        self._mol.construction_bonds.append(bond1)

        bond2 = Bond(boron_atom, diol.bonders[1], 1, periodicity)
        self._mol.bonds.append(bond2)
        self._mol.construction_bonds.append(bond2)

    def _react_ring_amine_with_ring_amine(
        self,
        reaction_key,
        func_groups,
        periodicity
    ):
        """
        Creates bonds between functional groups.

        Parameters
        ----------
        reaction_key : :class:`._ReactionKey`
            The key for the reaction.

        func_groups : :class:`list` of :class:`.FunctionalGroup`
            The functional groups from which deleter atoms should be
            removed.

        periodicity : :class:`tuple` of :class:`int`
            Specifies the periodicity of the bonds added by the
            reaction, which bridge the `func_groups`. See
            :attr:`.Bond.periodicity`.

        Returns
        -------
        None : :class:`NoneType`

        """

        self._remove_deleters(func_groups)
        fg1, fg2 = func_groups
        c1 = next(a for a in fg1.bonders if a.atomic_number == 6)
        n1 = next(a for a in fg1.bonders if a.atomic_number == 7)

        c2 = next(a for a in fg2.bonders if a.atomic_number == 6)
        n2 = next(a for a in fg2.bonders if a.atomic_number == 7)

        n1_coord = np.array(self._mol._position_matrix[n1.id])
        n2_coord = np.array(self._mol._position_matrix[n2.id])
        c1_coord = np.array(self._mol._position_matrix[c1.id])
        c2_coord = np.array(self._mol._position_matrix[c2.id])

        n_joiner = elements.C(len(self._mol.atoms))
        self._mol.atoms.append(n_joiner)
        n_joiner_coord = (n1_coord + n2_coord) / 2
        self._mol._position_matrix.append(n_joiner_coord)

        nh1 = elements.H(len(self._mol.atoms))
        self._mol.atoms.append(nh1)
        nh1_coord = n_joiner_coord + np.array([0, 0, 1])
        self._mol._position_matrix.append(nh1_coord)

        nh2 = elements.H(len(self._mol.atoms))
        self._mol.atoms.append(nh2)
        nh2_coord = n_joiner_coord + np.array([0, 0, -1])
        self._mol._position_matrix.append(nh2_coord)

        nc_joiner1 = elements.C(len(self._mol.atoms))
        self._mol.atoms.append(nc_joiner1)
        nc_joiner1_coord = (c1_coord + n2_coord) / 2
        self._mol._position_matrix.append(nc_joiner1_coord)

        nc1h1 = elements.H(len(self._mol.atoms))
        self._mol.atoms.append(nc1h1)
        nc1h1_coord = nc_joiner1_coord + np.array([0, 0, 1])
        self._mol._position_matrix.append(nc1h1_coord)

        nc1h2 = elements.H(len(self._mol.atoms))
        self._mol.atoms.append(nc1h2)
        nc1h2_coord = nc_joiner1_coord + np.array([0, 0, -1])
        self._mol._position_matrix.append(nc1h2_coord)

        nc_joiner2 = elements.C(len(self._mol.atoms))
        self._mol.atoms.append(nc_joiner2)
        nc_joiner2_coord = (c2_coord + n1_coord) / 2
        self._mol._position_matrix.append(nc_joiner2_coord)

        nc2h1 = elements.H(len(self._mol.atoms))
        self._mol.atoms.append(nc2h1)
        nc2h1_coord = nc_joiner2_coord + np.array([0, 0, 1])
        self._mol._position_matrix.append(nc2h1_coord)

        nc2h2 = elements.H(len(self._mol.atoms))
        self._mol.atoms.append(nc2h2)
        nc2h2_coord = nc_joiner2_coord + np.array([0, 0, -1])
        self._mol._position_matrix.append(nc2h2_coord)

        bonds = (
            Bond(n1, n_joiner, 1),
            Bond(n_joiner, n2, 1, periodicity),
            Bond(n_joiner, nh1, 1),
            Bond(n_joiner, nh2, 1),
            Bond(c1, nc_joiner1, 1),
            Bond(nc_joiner1, n2, 1, periodicity),
            Bond(nc_joiner1, nc1h1, 1),
            Bond(nc_joiner1, nc1h2, 1),
            Bond(nc_joiner2, c2, 1, periodicity),
            Bond(n1, nc_joiner2, 1),
            Bond(nc_joiner2, nc2h1, 1),
            Bond(nc_joiner2, nc2h2, 1)
        )
        for bond in bonds:
            self._mol.bonds.append(bond)
            self._mol.construction_bonds.append(bond)



"""
Covalent Organic Framework
==========================

For usage examples see :class:`.COF`.

#. :class:`Honeycomb`
#. :class:`Hexagonal`
#. :class:`Square`
#. :class:`Kagome`
#. :class:`LinkerlessHoneycomb`

"""

import numpy as np
import itertools as it

from .topology_graph import TopologyGraph, VertexData, Vertex, EdgeData
from ...utilities import vector_angle, flatten


class _COFVertexData(VertexData):
    """
    Holds data for a COF vertex.

    Attributes
    ----------
    id : :class:`int`
        The id of the vertex. Must match the index in
        :attr:`TopologyGraph.vertices`.

    position : :class:`numpy.ndarray`
        The position of the vertex.

    edges : :class:`list` of :class:`.EdgeData`
        The edges connected to the vertex.

    cell : :class:`numpy.ndarray`
        The unit cell in which the vertex is found.

    aligner_edge : :class:`int`
        The edge which is used to align the :class:`.BuildingBlock`
        placed on the vertex. The first :class:`.FunctionalGroup`
        in :attr:`.BuildingBlock.func_groups` is rotated such that
        it lies exactly on this :class:`.Edge`. Must be between
        ``0`` and the number of edges the vertex is connected to.

    """

    def __init__(self, x, y, z):
        """
        Initialize a :class:`.VertexData` instance.

        Parameters
        ----------
        x : :class:`float`
            The x coordinate.

        y : :class:`float`
            The y coordinate.

        z : :class:`float`
            The z coordinate.

        """

        self.aligner_edge = None
        super().__init__(x, y, z)

    @classmethod
    def init_at_center(cls, *vertex_data):
        obj = super().init_at_center(*vertex_data)
        obj.aligner_edge = None
        return obj

    @classmethod
    def init_at_shifted_center(
        cls,
        vertex_data,
        shifts,
        lattice_constants,
        aligner_edge=None
    ):
        """
        Initialize at the center of shifted `vertex_data`.

        Parameters
        ----------
        vertex_data : :class:`tuple` of :class:`.VertexData`
            The vertices at whose center this vertex should be
            intialized.

        shifts : :class:`tuple`
            For every vertex in `vertices` the amount by which it is
            shifted along each axis. For example

            .. code-block:: python

                shifts = (
                    (1, 0, -1),
                    (0, 0, 0)
                )

            means that the first vertex in `vertices` is shifted
            up along the x axis, is not shifted along the y axis
            and is shifted down along the z axis and the second
            vertex is not shifted at all.

        lattice_constants : :class:`tuple` of :class:`numpy.ndarray`
            The a, b and c lattice constants, each written as a vector.

        """

        positions = []
        for vertex, shift in zip(vertex_data, shifts):
            total_shift = 0
            for dim_shift, constant in zip(shift, lattice_constants):
                total_shift += dim_shift * constant
            positions.append(vertex.position + total_shift)

        position = np.divide(
            np.sum(positions, axis=0),
            len(positions)
        )
        return cls(*position)

    def clone(self, clear_edges=False):
        clone = super().clone(clear_edges)
        clone.aligner_edge = self.aligner_edge
        return clone

    def get_vertex(self):
        return _COFVertex(self)


class _COFVertex(Vertex):
    """
    Represents a vertex of a :class:`.COF`.

    Attributes
    ----------
    id : :class:`int`
        The id of the vertex. This should be its index in
        :attr:`.TopologyGraph.vertices`.

    """

    def __init__(self, data):
        """
        Initialize a :class:`_COFVertex`.

        Parameters
        ----------
        data : :class:`_COFVertexData`
            The vertex data.

        """

        # The edge which is used to align the :class:`.BuildingBlock`
        # placed on the vertex. The first :class:`.FunctionalGroup`
        # in :attr:`.BuildingBlock.func_groups` is rotated such that
        # it lies exactly on this :class:`.Edge`. Must be between
        # ``0`` and the number of edges the vertex is connected to.
        self._aligner_edge = data.aligner_edge
        super().__init__(data)

    def clone(self, clear_edges=False):
        """
        Return a clone.

        Parameters
        ----------
        clear_edges : :class:`bool`, optional
            ``True`` if the clone should not be connected to any edges.

        Returns
        -------
        :class:`Vertex`
            The clone.
        """

        clone = super().clone(clear_edges)
        clone._aligner_edge = self._aligner_edge
        return clone

    def get_aligner_edge(self):
        return self._aligner_edge

    def place_building_block(self, building_block, vertices, edges):
        """
        Place `building_block` on the :class:`.Vertex`.

        Parameters
        ----------
        building_block : :class:`.BuildingBlock`
            The building block molecule which is to be placed on the
            vertex.

        vertices : :class:`tuple` of :class:`.Vertex`
            All vertices in the topology graph. The index of each
            vertex must match its :class:`~.Vertex.id`.

        edges : :class:`tuple` of :class:`.Edge`
            All edges in the topology graph. The index of each
            edge must match its :class:`~.Edge.id`.

        Returns
        -------
        :class:`numpy.nadarray`
            The position matrix of `building_block` after being
            placed.

        """

        if len(building_block.func_groups) == 2:
            return self._place_linear_building_block(
                building_block=building_block,
                vertices=vertices,
                edges=edges
            )
        return self._place_nonlinear_building_block(
            building_block=building_block,
            vertices=vertices,
            edges=edges
        )

    def _place_linear_building_block(
        self,
        building_block,
        vertices,
        edges
    ):
        """
        Place `building_block` on the :class:`.Vertex`.

        Parameters
        ----------
        building_block : :class:`.BuildingBlock`
            The building block molecule which is to be placed on the
            vertex.

        vertices : :class:`tuple` of :class:`.Vertex`
            All vertices in the topology graph. The index of each
            vertex must match its :class:`~.Vertex.id`.

        edges : :class:`tuple` of :class:`.Edge`
            All edges in the topology graph. The index of each
            edge must match its :class:`~.Edge.id`.

        Returns
        -------
        :class:`numpy.nadarray`
            The position matrix of `building_block` after being
            placed.

        """

        building_block.set_centroid(
            position=self._position,
            atom_ids=building_block.get_bonder_ids()
        )
        fg_centroid = building_block.get_centroid(
            atom_ids=building_block.func_groups[0].get_bonder_ids()
        )
        start = fg_centroid - self._position
        e0_coord = (
            edges[self._edge_ids[0]].get_position(self, vertices)
        )
        e1_coord = (
            edges[self._edge_ids[1]].get_position(self, vertices)
        )
        target = e0_coord - e1_coord

        if self._edge_ids[self._aligner_edge] != self._edge_ids[0]:
            target *= -1

        building_block.apply_rotation_between_vectors(
            start=start,
            target=target,
            origin=self._position
        )
        start = building_block.get_centroid_centroid_direction_vector()
        building_block.apply_rotation_to_minimize_angle(
            start=start,
            target=self._position,
            axis=target,
            origin=self._position,
        )
        return building_block.get_position_matrix()

    def _place_nonlinear_building_block(
        self,
        building_block,
        vertices,
        edges
    ):
        """
        Place `building_block` on the :class:`.Vertex`.

        Parameters
        ----------
        building_block : :class:`.BuildingBlock`
            The building block molecule which is to be placed on the
            vertex.

        vertices : :class:`tuple` of :class:`.Vertex`
            All vertices in the topology graph. The index of each
            vertex must match its :class:`~.Vertex.id`.

        edges : :class:`tuple` of :class:`.Edge`
            All edges in the topology graph. The index of each
            edge must match its :class:`~.Edge.id`.

        Returns
        -------
        :class:`numpy.nadarray`
            The position matrix of `building_block` after being
            placed.

        """

        building_block.set_centroid(
            position=self._position,
            atom_ids=building_block.get_bonder_ids()
        )
        building_block.apply_rotation_between_vectors(
            start=building_block.get_bonder_plane_normal(),
            target=[0, 0, 1],
            origin=self._position
        )
        fg_bonder_centroid = building_block.get_centroid(
            atom_ids=building_block.func_groups[0].get_bonder_ids()
        )
        start = fg_bonder_centroid - self._position

        aligner_edge = edges[self._edge_ids[self._aligner_edge]]
        edge_coord = aligner_edge.get_position(self, vertices)
        target = edge_coord - self._position
        building_block.apply_rotation_to_minimize_angle(
            start=start,
            target=target,
            axis=[0, 0, 1],
            origin=self._position
        )
        return building_block.get_position_matrix()

    def assign_func_groups_to_edges(
        self,
        building_block,
        vertices,
        edges
    ):
        """
        Assign functional groups to edges.

        Each :class:`.FunctionalGroup` of the `building_block` needs
        to be associated with one of the :class:`.Edge` instances in
        :attr:`edges`.

        Parameters
        ----------
        building_block : :class:`.Molecule`
            The building block molecule which is needs to have
            functional groups assigned to edges.

        vertices : :class:`tuple` of :class:`.Vertex`
            All vertices in the topology graph. The index of each
            vertex must match its :class:`~.Vertex.id`.

        edges : :class:`tuple` of :class:`.Edge`
            All edges in the topology graph. The index of each
            edge must match its :class:`~.Edge.id`.

        Returns
        -------
        :class:`dict`
            A mapping from the id of a functional group in
            `building_block` to the id of the edge in :attr:`edges` it
            is assigned to.

        """

        if len(building_block.func_groups) == 2:
            return self._assign_func_groups_to_linear_edges(
                building_block=building_block,
                vertices=vertices,
                edges=edges
            )
        return self._assign_func_groups_to_nonlinear_edges(
                building_block=building_block,
                vertices=vertices,
                edges=edges
            )

    def _assign_func_groups_to_linear_edges(
        self,
        building_block,
        vertices,
        edges
    ):
        return {
            fg_id: edge_id for fg_id, edge_id in enumerate(sorted(
                self._edge_ids,
                key=self._get_fg0_distance(
                    building_block=building_block,
                    vertices=vertices,
                    edges=edges
                )
            ))
        }

    def _get_fg0_distance(self, building_block, vertices, edges):
        fg_coord = building_block.get_centroid(
            atom_ids=building_block.func_groups[0].get_bonder_ids()
        )

        def distance(edge_id):
            displacement = edges[edge_id].get_position(
                reference=self,
                vertices=vertices
            ) - fg_coord
            return np.linalg.norm(displacement)

        return distance

    def _assign_func_groups_to_nonlinear_edges(
        self,
        building_block,
        vertices,
        edges
    ):
        # The idea is to order the functional groups in building_block
        # by their angle from func_groups[0] and the bonder centroid,
        #  going in the clockwise direction.
        #
        # The edges are also ordered by their angle from aligner_edge
        # and the edge centroid going in the clockwise direction.
        #
        # Once the fgs and edges are ordered, zip and assign them.

        fg0_coord = building_block.get_centroid(
            atom_ids=building_block.func_groups[0].get_bonder_ids()
        )
        bonder_centroid = building_block.get_centroid(
            atom_ids=building_block.get_bonder_ids()
        )
        fg0_direction = fg0_coord-bonder_centroid
        axis = np.cross(
            fg0_direction,
            building_block.get_bonder_plane_normal()
        )
        func_groups = sorted(
            range(len(building_block.func_groups)),
            key=self._get_func_group_angle(
                building_block=building_block,
                fg0_direction=fg0_direction,
                bonder_centroid=bonder_centroid,
                axis=axis
            )
        )
        edge_ids = sorted(
            self._edge_ids,
            key=self._get_edge_angle(axis, vertices, edges)
        )
        assignments = {}
        for edge_id, fg_id in zip(edge_ids, func_groups):
            assignments[fg_id] = edge_id
        return assignments

    @staticmethod
    def _get_func_group_angle(
        building_block,
        fg0_direction,
        bonder_centroid,
        axis
    ):

        def angle(fg_id):
            func_group = building_block.func_groups[fg_id]
            coord = building_block.get_centroid(
                atom_ids=func_group.get_bonder_ids()
            )
            fg_direction = coord-bonder_centroid
            theta = vector_angle(fg0_direction, fg_direction)

            projection = fg_direction @ axis
            if theta > 0 and projection < 0:
                return 2*np.pi - theta
            return theta

        return angle

    def _get_edge_angle(self, axis, vertices, edges):

        aligner_edge = edges[self._edge_ids[self._aligner_edge]]
        aligner_edge_coord = aligner_edge.get_position(self, vertices)
        connected_edges = tuple(edges[id_] for id_ in self._edge_ids)
        edge_centroid = self._get_edge_centroid(
            centroid_edges=connected_edges,
            vertices=vertices
        )
        # This axis is used to figure out the clockwise direction.
        aligner_edge_direction = aligner_edge_coord - edge_centroid

        def angle(edge_id):
            coord = edges[edge_id].get_position(self, vertices)
            edge_direction = coord - edge_centroid
            theta = vector_angle(
                vector1=edge_direction,
                vector2=aligner_edge_direction
            )

            projection = edge_direction @ axis
            if theta > 0 and projection < 0:
                return 2*np.pi - theta
            return theta

        return angle

    def __str__(self):
        return (
            f'Vertex(id={self.id}, '
            f'position={self._position.tolist()}, '
            f'aligner_edge={self._aligner_edge})'
        )


class COF(TopologyGraph):
    """
    Represents a COF topology graph.

    COF topologies are added by creating a subclass which defines the
    :attr:`vertices` and :attr:`edges` of the topology as class
    attributes.

    Attributes
    ----------
    vertex_data : :class:`tuple` of :class:`.VertexData`
        A class attribute. Holds the data of the vertices which make up
        the topology graph.

    edge_data : :class:`tuple` of :class:`.EdgeData`
        A class attribute. Holds the data of the edges which make up
        the topology graph.

    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    Examples
    --------
    :class:`COF` instances can be made by supplying only
    the lattice size (using :class:`.Honeycomb` as an example)

    .. code-block:: python

        import stk

        bb1 = stk.BuildingBlock('NCCN', ['amine'])
        bb2 = stk.BuildingBlock('O=CC(C=O)C=O', ['aldehyde'])
        cof1 = stk.ConstructedMolecule(
            building_blocks=[bb1, bb2],
            topology_graph=stk.cof.Honeycomb((2, 2, 1))
        )

    Different structural isomers of COFs can be made by using the
    `vertex_alignments` optional parameter

    .. code-block:: python

        lattice = stk.cof.Honeycomb(
            lattice_size=(2, 2, 1),
            vertex_alignments={0: 1, 2: 2}
        )
        cof2 = stk.ConstructedMolecule(
            building_blocks=[bb1, bb2],
            topology_graph=lattice
        )

    The parameter maps the :attr:`~.Vertex.id` of a vertex to a number
    between 0 (inclusive) and the number of edges the vertex is
    connected to (exclusive). So a vertex connected to three edges
    can be mapped to ``0``, ``1`` or ``2``.

    You can also build COFs with multiple building blocks, but you
    have to assign each building block to a vertex with
    `building_block_vertices`.

    .. code-block:: python

        lattice = stk.cof.Honeycomb(
            lattice_size=(2, 2, 1),
            vertex_alignments={0: 1, 2: 2}
        )
        bb3 = stk.BuildingBlock('NCOCN', ['amine'])
        cof2 = stk.ConstructedMolecule(
            building_blocks=[bb1, bb2, bb3],
            topology_graph=lattice
            building_block_vertices={
                bb1: lattice.verices[:2],
                bb2: lattice.verices[4:],
                bb3: lattice.verices[2:4]
            }
        )

    """

    def __init_subclass__(cls, **kwargs):
        for i, vertex in enumerate(cls.vertex_data):
            vertex.id = i
        for i, edge in enumerate(cls.edge_data):
            edge.id = i
            edge.lattice_constants = tuple(
                np.array(constant)
                for constant in cls._lattice_constants
            )
        return super().__init_subclass__(**kwargs)

    def __init__(
        self,
        lattice_size,
        periodic=False,
        vertex_alignments=None,
        num_processes=1
    ):
        """
        Initialize a :class:`.COF`.

        Parameters
        ----------
        lattice_size : :class:`tuple` of :class:`int`
            The number of unit cells which should be placed along the
            x, y and z dimensions, respectively.

        periodic : :class:`bool`, optional
            If periodic bonds are to be made across the lattice,
            this should be ``True``. If ``False`` the functional
            groups on the ends of the lattice will be unreacted.

        vertex_alignments : :class:`dict`, optional
            A mapping from the :attr:`.Vertex.id` of a :class:`.Vertex`
            :attr:`vertices` to an :class:`.Edge` connected to it.
            The :class:`.Edge` is used to align the first
            :class:`.FunctionalGroup` of a :class:`.BuildingBlock`
            placed on that vertex. Only vertices which need to have
            their default edge changed need to be present in the
            :class:`dict`. If ``None`` then the default edge is used
            for each vertex. Changing which :class:`.Edge` is used will
            mean that the topology graph represents different
            structural isomers. The edge is refered to by a number
            between ``0`` (inclusive) and the number of edges the
            vertex is connected to (exclusive).

        num_processes : :class:`int`, optional
            The number of parallel processes to create during
            :meth:`construct`.

        """

        if vertex_alignments is None:
            vertex_alignments = {}

        self._lattice_size = lattice_size
        self._periodic = periodic

        vertex_data = self._get_vertex_data(vertex_alignments)
        edge_data = self._get_edge_data(vertex_data)

        vertex_data = tuple(
            vertex
            for clones in flatten(vertex_data, {dict})
            for vertex in clones.values()
        )
        super().__init__(vertex_data, edge_data, (), num_processes)

    def _get_vertex_data(self, vertex_alignments):
        """
        Create the vertex data of the topology graph instance.

        Parameters
        ---------
        vertex_alignments : :class:`dict`
            A mapping from the :attr:`.Vertex.id` of a :class:`.Vertex`
            :attr:`vertices` to an :class:`.Edge` connected to it.
            The :class:`.Edge` is used to align the first
            :class:`.FunctionalGroup` of a :class:`.BuildingBlock`
            placed on that vertex. Only vertices which need to have
            their default edge changed need to be present in the
            :class:`dict`. If ``None`` then the default edge is used
            for each vertex. Changing which :class:`.Edge` is used will
            mean that the topology graph represents different
            structural isomers. The edge is refered to by a number
            between ``0`` (inclusive) and the number of edges the
            vertex is connected to (exclusive).

        Returns
        -------
        :class:`list`
            A nested :class:`list` which can be indexed as
            ``vertices[x][y][z]``, which will return a :class:`dict`
            for the unit cell at (x, y, z). The :class:`dict` maps
            the vertices in :attr:`vertex_data` to its clone for that
            unit cell.

        """

        xdim, ydim, zdim = (range(dim) for dim in self._lattice_size)
        # vertex_clones is indexed as vertex_clones[x][y][z]
        vertex_clones = [
            [
                [
                    {} for k in zdim
                ]
                for j in ydim
            ]
            for i in xdim
        ]
        # Make a clone of each vertex for each unit cell.
        cells = it.product(xdim, ydim, zdim)
        vertices = it.product(cells, self.vertex_data)
        for cell, vertex in vertices:
            x, y, z = cell
            clone = vertex.clone(True)
            clone.cell = np.array(cell)
            clone.aligner_edge = vertex_alignments.get(vertex.id, 0)
            # Shift the clone so that it's within the cell.
            for axis, dim in zip(cell, self._lattice_constants):
                clone.position += axis * dim

            vertex_clones[x][y][z][vertex] = clone
        return vertex_clones

    def _get_edge_data(self, vertex_data):
        """
        Create the edge data of the topology graph instance.

        Parameters
        ----------
        vertex_data : :class:`list`
            A nested :class:`list` which can be indexed as
            ``vertex_data[x][y][z]``, which will return a :class:`dict`
            for the unit cell at (x, y, z). The :class:`dict` maps
            the vertices in :attr:`vertex_data` to the clones for that
            unit cell.

        Returns
        -------
        :class:`tuple` of :class:`.EdgeData`
            The edge data of the topology graph instance.

        """

        edge_clones = []
        # Make a clone for each edge for each unit cell.
        xdim, ydim, zdim = (range(dim) for dim in self._lattice_size)
        cells = it.product(xdim, ydim, zdim)
        edges = it.product(cells, self.edge_data)
        for cell, edge in edges:
            x, y, z = cell
            # The cell in which the second vertex of the edge is found.
            periodic_cell = np.array(cell) + edge.periodicity
            # Wrap around periodic cells, ie those that are less than 0
            # or greater than the lattice size along any dimension.
            dims = zip(periodic_cell, self._lattice_size)
            cell2_x, cell2_y, cell2_z = np.array([
                (dim+max_dim) % max_dim
                for dim, max_dim in dims
            ])
            # Make a vertex map which accounts for the fact that
            # v1 is in cell2.
            v0, v1 = edge.vertices
            vertex_map = {
                v0: vertex_data[x][y][z][v0],
                v1: vertex_data[cell2_x][cell2_y][cell2_z][v1]
            }
            # If the edge is not periodic if periodic_cell is did not
            # have to wrap around.
            dims = zip(periodic_cell, self._lattice_size)
            edge_is_not_periodic = all(
                dim >= 0 and dim < max_dim
                for dim, max_dim in dims
            )
            clone = edge.clone(vertex_map, True, True)
            edge_clones.append(clone)
            if edge_is_not_periodic:
                clone.periodicity = np.array([0, 0, 0])

        return tuple(edge_clones)

    def _before_react(self, mol, vertices, edges):
        if self._periodic:
            return vertices, edges
        return vertices, tuple(
            edge for edge in edges if not edge.is_periodic()
        )

    def assign_building_blocks_to_vertices(self, building_blocks):
        """
        Assign `building_blocks` to :attr:`vertices`.

        Parameters
        ----------
        building_blocks : :class:`list` of :class:`.Molecule`
            The :class:`.BuildingBlock` and
            :class:`ConstructedMolecule` instances which
            represent the building block molecules used for
            construction. Only one instance is present per building
            block molecule, even if multiples of that building block
            join up to form the :class:`ConstructedMolecule`.

        Returns
        -------
        :class:`dict`
            Maps the `building_blocks`, to the
            :class:`~.topologies.base.Vertex` objects in
            :attr:`vertices` they are placed on during construction.
            The :class:`dict` has the form

            .. code-block:: python

                building_block_vertices = {
                    BuildingBlock(...): [Vertex(...), Vertex(...)],
                    BuildingBlock(...): [
                        Vertex(...),
                        Vertex(...),
                        Vertex(...),
                    ]
                    ConstructedMolecule(...): [Vertex(...)]
                }

        Raises
        ------
        :class:`ValueError`
            If there is more than one building with a given number
            of functional groups.

        """

        bb_by_degree = {}
        for bb in building_blocks:
            num_fgs = len(bb.func_groups)
            if num_fgs in bb_by_degree:
                raise ValueError(
                    'If there are multiple building blocks with the '
                    'same number of functional groups, '
                    'building_block_vertices must be set explicitly.'
                )
            bb_by_degree[num_fgs] = bb

        building_block_vertices = {}
        for vertex in self.vertices:
            bb = bb_by_degree[vertex.get_num_edges()]
            building_block_vertices[bb] = (
                building_block_vertices.get(bb, [])
            )
            building_block_vertices[bb].append(vertex)
        return building_block_vertices

    def _get_scale(self, mol):
        """
        Get the scale used for the positions of :attr:`vertices`.

        Parameters
        ----------
        mol : :class:`.ConstructedMolecule`
            The molecule being constructed.

        Returns
        -------
        :class:`float` or :class:`list` of :class:`float`
            The value by which the position of each :class:`Vertex` is
            scaled. Can be a single number if all axes are scaled by
            the same amount or a :class:`list` of three numbers if
            each axis is scaled by a different value.

        """

        return 5*max(
            bb.get_maximum_diameter()
            for bb in mol.building_block_vertices
        )

    def __repr__(self):
        vertex_alignments = ', '.join(
            f'{v.id}: {v.get_aligner_edge()}'
            # Only get the vertices in the first unit cell.
            for v in self.vertices[:len(self.vertex_data)]
        )

        x, y, z = self._lattice_size
        periodic = ', periodic=True' if self._periodic else ''
        return (
            f'cof.{self.__class__.__name__}('
            f'lattice_size=({x}, {y}, {z}), '
            f'vertex_alignments={{{vertex_alignments}}}'
            f'{periodic})'
        )


class Honeycomb(COF):
    """
    Represents a honeycomb COF topology graph.

    Building blocks with three and two functional groups are required
    for this topology graph.

    See :class:`.COF` for more details and examples.

    Attributes
    ----------
    vertex_data : :class:`tuple` of :class:`.VertexData`
        A class attribute. Holds the data of the vertices which make up
        the topology graph.

    edge_data : :class:`tuple` of :class:`.EdgeData`
        A class attribute. Holds the data of the edges which make up
        the topology graph.

    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    """

    _lattice_constants = _a, _b, _c = (
        np.array([1., 0., 0.]),
        np.array([0.5, 0.866, 0]),
        np.array([0, 0, 5/1.7321])
    )

    _vertex_data = (
        _COFVertexData(*((1/3)*_a + (1/3)*_b + (1/2)*_c)),
        _COFVertexData(*((2/3)*_a + (2/3)*_b + (1/2)*_c))
    )

    vertex_data = (
        *_vertex_data,
        _COFVertexData.init_at_center(
            _vertex_data[0], _vertex_data[1]
        ),
        _COFVertexData.init_at_shifted_center(
            vertex_data=(_vertex_data[0], _vertex_data[1]),
            shifts=((0, 0, 0), (0, -1, 0)),
            lattice_constants=_lattice_constants
        ),
        _COFVertexData.init_at_shifted_center(
            vertex_data=(_vertex_data[0], _vertex_data[1]),
            shifts=((0, 0, 0), (-1, 0, 0)),
            lattice_constants=_lattice_constants
        )
    )

    edge_data = (
        EdgeData(vertex_data[2], vertex_data[0]),
        EdgeData(vertex_data[2], vertex_data[1]),

        EdgeData(vertex_data[3], vertex_data[0]),
        EdgeData(
            vertex_data[3],
            vertex_data[1],
            periodicity=(0, -1, 0)
        ),

        EdgeData(vertex_data[4], vertex_data[0]),
        EdgeData(
            vertex_data[4],
            vertex_data[1],
            periodicity=(-1, 0, 0)
        )
    )


class Hexagonal(COF):
    """
    Represents a hexagonal COF topology graph.

    Building blocks with six and two functional groups are required
    for this topology graph.

    See :class:`.COF` for more details and examples.

    Attributes
    ----------
    vertex_data : :class:`tuple` of :class:`.VertexData`
        A class attribute. Holds the data of the vertices which make up
        the topology graph.

    edge_data : :class:`tuple` of :class:`.EdgeData`
        A class attribute. Holds the data of the edges which make up
        the topology graph.

    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    """

    _lattice_constants = _a, _b, _c = (
        np.array([1., 0., 0.]),
        np.array([0.5, 0.866, 0]),
        np.array([0, 0, 5/1.7321])
    )

    _vertex_data = (
        _COFVertexData(*((1/4)*_a + (1/4)*_b + (1/2)*_c)),
        _COFVertexData(*((1/4)*_a + (3/4)*_b + (1/2)*_c)),
        _COFVertexData(*((3/4)*_a + (1/4)*_b + (1/2)*_c)),
        _COFVertexData(*((3/4)*_a + (3/4)*_b + (1/2)*_c))
    )

    vertex_data = (
        *_vertex_data,
        _COFVertexData.init_at_center(
            _vertex_data[0], _vertex_data[1]
        ),
        _COFVertexData.init_at_center(
            _vertex_data[0], _vertex_data[2]
        ),
        _COFVertexData.init_at_center(
            _vertex_data[1], _vertex_data[2]
        ),
        _COFVertexData.init_at_center(
            _vertex_data[1], _vertex_data[3]
        ),
        _COFVertexData.init_at_center(
            _vertex_data[2], _vertex_data[3]
        ),
        _COFVertexData.init_at_shifted_center(
            vertex_data=(_vertex_data[0], _vertex_data[2]),
            shifts=((0, 0, 0), (-1, 0, 0)),
            lattice_constants=_lattice_constants
        ),
        _COFVertexData.init_at_shifted_center(
            vertex_data=(_vertex_data[0], _vertex_data[1]),
            shifts=((0, 0, 0), (0, -1, 0)),
            lattice_constants=_lattice_constants
        ),
        _COFVertexData.init_at_shifted_center(
            vertex_data=(_vertex_data[0], _vertex_data[3]),
            shifts=((0, 0, 0), (0, -1, 0)),
            lattice_constants=_lattice_constants
        ),
        _COFVertexData.init_at_shifted_center(
            vertex_data=(_vertex_data[2], _vertex_data[1]),
            shifts=((0, 0, 0), (1, -1, 0)),
            lattice_constants=_lattice_constants
        ),
        _COFVertexData.init_at_shifted_center(
            vertex_data=(_vertex_data[2], _vertex_data[3]),
            shifts=((0, 0, 0), (0, -1, 0)),
            lattice_constants=_lattice_constants
        ),
        _COFVertexData.init_at_shifted_center(
            vertex_data=(_vertex_data[1], _vertex_data[3]),
            shifts=((0, 0, 0), (-1, 0, 0)),
            lattice_constants=_lattice_constants
        ),
        _COFVertexData.init_at_shifted_center(
            vertex_data=(_vertex_data[3], _vertex_data[0]),
            shifts=((0, 0, 0), (1, 0, 0)),
            lattice_constants=_lattice_constants
        )
    )

    edge_data = (
        EdgeData(vertex_data[4], vertex_data[0]),
        EdgeData(vertex_data[4], vertex_data[1]),

        EdgeData(vertex_data[5], vertex_data[0]),
        EdgeData(vertex_data[5], vertex_data[2]),

        EdgeData(vertex_data[6], vertex_data[1]),
        EdgeData(vertex_data[6], vertex_data[2]),

        EdgeData(vertex_data[7], vertex_data[1]),
        EdgeData(vertex_data[7], vertex_data[3]),

        EdgeData(vertex_data[8], vertex_data[2]),
        EdgeData(vertex_data[8], vertex_data[3]),

        EdgeData(vertex_data[9], vertex_data[0]),
        EdgeData(
            vertex_data[9],
            vertex_data[2],
            periodicity=(-1, 0, 0)
        ),

        EdgeData(vertex_data[10], vertex_data[0]),
        EdgeData(
            vertex_data[10],
            vertex_data[1],
            periodicity=(0, -1, 0)
        ),

        EdgeData(vertex_data[11], vertex_data[0]),
        EdgeData(
            vertex_data[11],
            vertex_data[3],
            periodicity=(0, -1, 0)
        ),

        EdgeData(vertex_data[12], vertex_data[2]),
        EdgeData(
            vertex_data[12],
            vertex_data[1],
            periodicity=(1, -1, 0)
        ),

        EdgeData(vertex_data[13], vertex_data[2]),
        EdgeData(
            vertex_data[13],
            vertex_data[3],
            periodicity=(0, -1, 0)
        ),

        EdgeData(vertex_data[14], vertex_data[1]),
        EdgeData(
            vertex_data[14],
            vertex_data[3],
            periodicity=(-1, 0, 0)
        ),

        EdgeData(vertex_data[15], vertex_data[3]),
        EdgeData(
            vertex_data[15],
            vertex_data[0],
            periodicity=(1, 0, 0)
        )
    )


class Square(COF):
    """
    Represents a sqaure COF topology graph.

    Building blocks with four and two functional groups are required
    for this topology graph.

    See :class:`.COF` for more details and examples.

    Attributes
    ----------
    vertex_data : :class:`tuple` of :class:`.VertexData`
        A class attribute. Holds the data of the vertices which make up
        the topology graph.

    edge_data : :class:`tuple` of :class:`.EdgeData`
        A class attribute. Holds the data of the edges which make up
        the topology graph.

    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    """

    _lattice_constants = _a, _b, _c = (
        np.array([1., 0., 0.]),
        np.array([0., 1., 0.]),
        np.array([0., 0., 1.])
    )

    _vertex_data = (
        _COFVertexData(*((0.5)*_a + (0.5)*_b + (0.5)*_c)),
    )
    vertex_data = (
        *_vertex_data,
        _COFVertexData.init_at_shifted_center(
            vertex_data=(_vertex_data[0], _vertex_data[0]),
            shifts=((0, 0, 0), (1, 0, 0)),
            lattice_constants=_lattice_constants
        ),
        _COFVertexData.init_at_shifted_center(
            vertex_data=(_vertex_data[0], _vertex_data[0]),
            shifts=((0, 0, 0), (0, 1, 0)),
            lattice_constants=_lattice_constants
        )

    )

    edge_data = (
        EdgeData(vertex_data[1], vertex_data[0]),
        EdgeData(
            vertex_data[1],
            vertex_data[0],
            periodicity=(1, 0, 0)
        ),
        EdgeData(vertex_data[2], vertex_data[0]),
        EdgeData(
            vertex_data[2],
            vertex_data[0],
            periodicity=(0, 1, 0)
        )
    )


class Kagome(COF):
    """
    Represents a kagome COF topology graph.

    Building blocks with four and two functional groups are required
    for this topology graph.

    See :class:`.COF` for more details and examples.

    Attributes
    ----------
    vertex_data : :class:`tuple` of :class:`.VertexData`
        A class attribute. Holds the data of the vertices which make up
        the topology graph.

    edge_data : :class:`tuple` of :class:`.EdgeData`
        A class attribute. Holds the data of the edges which make up
        the topology graph.

    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    """

    _lattice_constants = _a, _b, _c = (
        np.array([1., 0., 0.]),
        np.array([0.5, 0.866, 0.]),
        np.array([0., 0., 5/1.7321])
    )

    _vertex_data = (
        _COFVertexData(*((1/4)*_a + (3/4)*_b + (0.5)*_c)),
        _COFVertexData(*((3/4)*_a + (3/4)*_b + (1/2)*_c)),
        _COFVertexData(*((3/4)*_a + (1/4)*_b + (1/2)*_c))
    )

    vertex_data = (
        *_vertex_data,
        _COFVertexData.init_at_center(
            _vertex_data[0],
            _vertex_data[1]
        ),
        _COFVertexData.init_at_center(
            _vertex_data[0],
            _vertex_data[2]
        ),
        _COFVertexData.init_at_center(
            _vertex_data[1],
            _vertex_data[2]
        ),
        _COFVertexData.init_at_shifted_center(
            vertex_data=(_vertex_data[0], _vertex_data[1]),
            shifts=((0, 0, 0), (-1, 0, 0)),
            lattice_constants=_lattice_constants
        ),
        _COFVertexData.init_at_shifted_center(
            vertex_data=(_vertex_data[0], _vertex_data[2]),
            shifts=((0, 0, 0), (-1, 1, 0)),
            lattice_constants=_lattice_constants
        ),
        _COFVertexData.init_at_shifted_center(
            vertex_data=(_vertex_data[1], _vertex_data[2]),
            shifts=((0, 0, 0), (0, 1, 0)),
            lattice_constants=_lattice_constants
        )

    )

    edge_data = (
        EdgeData(vertex_data[3], vertex_data[0]),
        EdgeData(vertex_data[3], vertex_data[1]),

        EdgeData(vertex_data[4], vertex_data[0]),
        EdgeData(vertex_data[4], vertex_data[2]),

        EdgeData(vertex_data[5], vertex_data[1]),
        EdgeData(vertex_data[5], vertex_data[2]),

        EdgeData(vertex_data[6], vertex_data[0]),
        EdgeData(
            vertex_data[6],
            vertex_data[1],
            periodicity=(-1, 0, 0)
        ),

        EdgeData(vertex_data[7], vertex_data[0]),
        EdgeData(
            vertex_data[7],
            vertex_data[2],
            periodicity=(-1, 1, 0)
        ),

        EdgeData(vertex_data[8], vertex_data[1]),
        EdgeData(
            vertex_data[8],
            vertex_data[2],
            periodicity=(0, 1, 0)
        )
    )


class LinkerlessHoneycomb(COF):
    """
    Represents a honeycomb COF topology graph.

    Building blocks with three functional groups are required
    for this topology graph.

    See :class:`.COF` for more details and examples.

    Attributes
    ----------
    vertex_data : :class:`tuple` of :class:`.VertexData`
        A class attribute. Holds the data of the vertices which make up
        the topology graph.

    edge_data : :class:`tuple` of :class:`.EdgeData`
        A class attribute. Holds the data of the edges which make up
        the topology graph.

    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    """

    _lattice_constants = _a, _b, _c = (
        np.array([1., 0., 0.]),
        np.array([0.5, 0.866, 0.]),
        np.array([0., 0., 5/1.7321])
    )

    vertex_data = (
        _COFVertexData(*((1/3)*_a + (1/3)*_b + (1/2)*_c)),
        _COFVertexData(*((2/3)*_a + (2/3)*_b + (1/2)*_c))
    )

    edge_data = (
        EdgeData(vertex_data[0], vertex_data[1]),
        EdgeData(
            vertex_data[0],
            vertex_data[1],
            periodicity=(-1, 0, 0)
        ),
        EdgeData(
            vertex_data[0],
            vertex_data[1],
            periodicity=(0, -1, 0)
        )
    )



"""
Defines cages made from building blocks with 2 and 5 functional groups.

"""

from scipy.constants import golden

from .base import Cage, _CageVertexData
from ..topology_graph import EdgeData


class TwelvePlusThirty(Cage):
    """
    Represents a icosahedron cage topology graph.

    See :class:`.Cage` for more details and examples.

    Attributes
    ----------
    vertex_data : :class:`tuple` of :class:`.VertexData`
        A class attribute. Holds the data of the vertices which make up
        the topology graph.

    edge_data : :class:`tuple` of :class:`.EdgeData`
        A class attribute. Holds the data of the edges which make up
        the topology graph.

    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    """

    # Vertices of a regular origin-centred icosahedron
    # Source: http://eusebeia.dyndns.org/4d/icosahedron
    _vertex_data = (
        _CageVertexData(0, 1, golden),
        _CageVertexData(0, -1, golden),
        _CageVertexData(0, 1, -golden),
        _CageVertexData(0, -1, -golden),
        _CageVertexData(1, golden, 0),
        _CageVertexData(-1, golden, 0),
        _CageVertexData(1, -golden, 0),
        _CageVertexData(-1, -golden, 0),
        _CageVertexData(golden, 0, 1),
        _CageVertexData(-golden, 0, 1),
        _CageVertexData(golden, 0, -1),
        _CageVertexData(-golden, 0, -1)
    )

    vertex_data = (
        *_vertex_data,
        _CageVertexData.init_at_center(
            _vertex_data[0], _vertex_data[1]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[0], _vertex_data[9]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[0], _vertex_data[5]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[0], _vertex_data[4]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[0], _vertex_data[8]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[8], _vertex_data[1]),
        _CageVertexData.init_at_center(
            _vertex_data[1], _vertex_data[9]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[9], _vertex_data[5]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[5], _vertex_data[4]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[4], _vertex_data[8]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[5], _vertex_data[2]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[5], _vertex_data[11]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[9], _vertex_data[11]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[9], _vertex_data[7]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[1], _vertex_data[7]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[1], _vertex_data[6]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[8], _vertex_data[6]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[8], _vertex_data[10]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[4], _vertex_data[10]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[4], _vertex_data[2]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[2], _vertex_data[11]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[11], _vertex_data[7]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[7], _vertex_data[6]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[6], _vertex_data[10]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[10], _vertex_data[2]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[2], _vertex_data[3]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[11], _vertex_data[3]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[7], _vertex_data[3]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[6], _vertex_data[3]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[10], _vertex_data[3]
        )
    )

    edge_data = (
        EdgeData(vertex_data[12], vertex_data[0]),
        EdgeData(vertex_data[12], vertex_data[1]),
        EdgeData(vertex_data[13], vertex_data[0]),
        EdgeData(vertex_data[13], vertex_data[9]),

        EdgeData(vertex_data[14], vertex_data[0]),
        EdgeData(vertex_data[14], vertex_data[5]),
        EdgeData(vertex_data[15], vertex_data[0]),
        EdgeData(vertex_data[15], vertex_data[4]),

        EdgeData(vertex_data[16], vertex_data[0]),
        EdgeData(vertex_data[16], vertex_data[8]),
        EdgeData(vertex_data[17], vertex_data[8]),
        EdgeData(vertex_data[17], vertex_data[1]),

        EdgeData(vertex_data[18], vertex_data[1]),
        EdgeData(vertex_data[18], vertex_data[9]),
        EdgeData(vertex_data[19], vertex_data[9]),
        EdgeData(vertex_data[19], vertex_data[5]),

        EdgeData(vertex_data[20], vertex_data[5]),
        EdgeData(vertex_data[20], vertex_data[4]),
        EdgeData(vertex_data[21], vertex_data[4]),
        EdgeData(vertex_data[21], vertex_data[8]),

        EdgeData(vertex_data[22], vertex_data[5]),
        EdgeData(vertex_data[22], vertex_data[2]),
        EdgeData(vertex_data[23], vertex_data[5]),
        EdgeData(vertex_data[23], vertex_data[11]),

        EdgeData(vertex_data[24], vertex_data[9]),
        EdgeData(vertex_data[24], vertex_data[11]),
        EdgeData(vertex_data[25], vertex_data[9]),
        EdgeData(vertex_data[25], vertex_data[7]),

        EdgeData(vertex_data[26], vertex_data[1]),
        EdgeData(vertex_data[26], vertex_data[7]),
        EdgeData(vertex_data[27], vertex_data[1]),
        EdgeData(vertex_data[27], vertex_data[6]),

        EdgeData(vertex_data[28], vertex_data[8]),
        EdgeData(vertex_data[28], vertex_data[6]),
        EdgeData(vertex_data[29], vertex_data[8]),
        EdgeData(vertex_data[29], vertex_data[10]),

        EdgeData(vertex_data[30], vertex_data[4]),
        EdgeData(vertex_data[30], vertex_data[10]),
        EdgeData(vertex_data[31], vertex_data[4]),
        EdgeData(vertex_data[31], vertex_data[2]),

        EdgeData(vertex_data[32], vertex_data[2]),
        EdgeData(vertex_data[32], vertex_data[11]),
        EdgeData(vertex_data[33], vertex_data[11]),
        EdgeData(vertex_data[33], vertex_data[7]),

        EdgeData(vertex_data[34], vertex_data[7]),
        EdgeData(vertex_data[34], vertex_data[6]),
        EdgeData(vertex_data[35], vertex_data[6]),
        EdgeData(vertex_data[35], vertex_data[10]),

        EdgeData(vertex_data[36], vertex_data[10]),
        EdgeData(vertex_data[36], vertex_data[2]),
        EdgeData(vertex_data[37], vertex_data[2]),
        EdgeData(vertex_data[37], vertex_data[3]),

        EdgeData(vertex_data[38], vertex_data[11]),
        EdgeData(vertex_data[38], vertex_data[3]),
        EdgeData(vertex_data[39], vertex_data[7]),
        EdgeData(vertex_data[39], vertex_data[3]),

        EdgeData(vertex_data[40], vertex_data[6]),
        EdgeData(vertex_data[40], vertex_data[3]),
        EdgeData(vertex_data[41], vertex_data[10]),
        EdgeData(vertex_data[41], vertex_data[3])
    )

    num_windows = 20
    num_window_types = 1



"""
Adding Topology Graphs
======================

To add a new topology graph a new subclass of :class:`.TopologyGraph`
must be added, which implements its virtual methods. Similarly, new
subclasses of :class:`.VertexData`` and :class:`.Vertex` must also be
made and their virtual methods implemented. When the new subclass of
:class:`.TopologyGraph` is initialized, it must create instances of the
:class:`.VertexData`
subclass, together with :class:`.EdgeData` instances. Once your
topology graph has the vertex and edge data it wants, simply run
:meth:`.TopologyGraph.__init__` and you're done.

When creating a :class:`.VertexData` subclass,
:meth:`~.VertexData.get_vertex` needs to be implemented such that it
returns an instance of your :class:`.Vertex` subclass. If you need to
define a new :meth:`__init__` method for either subclass, you will
also need to implement :meth:`clone` for it.

The :class:`.TopologyGraph` subclass can also create
`construction_stages` if parallel construction needs to be broken down
into separate stages. However,
if this is not the case, then an empty :class:`tuple` can simply be
passed.

Why is both :class:`.VertexData` and :class:`.Vertex` needed?
-------------------------------------------------------------

At first, it may appear that having both :class:`.VertexData` and
:class:`.Vertex` is an unnecssary inconvenience, as when you create
a new :class:`.TopologyGraph` subclass you have to subclass both of
these classes rather than just :class:`.Vertex`. The answer is
related to how these two classes reference other objects in the
:class:`.TopologyGraph`.

:class:`.VertexData` and :class:`.EdgeData` objects keep pointers
to each other in the :attr:`~.VertexData.edges` and
:attr:`~.EdgeData.vertices`. This is extremely convenient for
defining a :class:`.TopologyGraph` because its components can directly
reference each other. However, it poses a significant issue for
serialization. Topology graphs are usually highly-cyclic structures
and are therefore often not possible to serialize with off-the-shelf
serialization tools like :mod:`pickle` or :mod:`dill`. However,
serialization is necessary and fundamental for allowing
parallelization of :class:`.TopologyGraph` construction. The
vertices and edges of the graph have to be serialized and sent to
other cores so that they can place and connect building blocks in
parallel. As a  result, :class:`.VertexData` exists to allow a
convenient definition of a :class:`TopologyGraph`, while
:class:`.Vertex` exists to provide a serializable representation of it.
:class:`.Verex` and :class:`Edge` do not reference other objects
directly, instead they refer to them by their :attr:`.Vertex.id`,
which is used to get an index into :attr:`.TopologyGraph.vertices`
and :attr:`.TopologyGraph.edges`.

:meth:`.VertexData.get_vertex` is used to convert :class:`.VertexData`
into its :class:`.Vertex` counterpart.

"""

import numpy as np
import pathos
from collections import namedtuple

from ..reactor import Reactor
from ...utilities import vector_angle


class VertexData:
    """
    Holds the data used to initialize a :class:`.Vertex`.

    Attributes
    ----------
    id : :class:`int`
        The id of the vertex. Must match the index in
        :attr:`TopologyGraph.vertices`.

    position : :class:`numpy.ndarray`
        The position of the vertex.

    edges : :class:`list` of :class:`.EdgeData`
        The edges connected to the vertex.

    cell : :class:`numpy.ndarray`
        The unit cell in which the vertex is found.

    """

    def __init__(self, x, y, z):
        """
        Initialize a :class:`.VertexData` instance.

        Parameters
        ----------
        x : :class:`float`
            The x coordinate.

        y : :class:`float`
            The y coordinate.

        z : :class:`float`
            The z coordinate.

        """

        # Set by TopologyGraph.__init__.
        self.id = None
        self.position = np.array([x, y, z], dtype=np.dtype('float64'))
        self.edges = []
        self.cell = np.array([0, 0, 0])

    @classmethod
    def init_at_center(cls, *vertex_data):
        """
        Initialize at the center of other vertices.

        Parameters
        ----------
        *vertex_data : :class:`.VertexData`
            Vertices at whose center this vertex should be initialized.

        Returns
        -------
        :class:`.VertexData`
            The vertex.

        """

        center = sum(vertex.position for vertex in vertex_data)
        center /= len(vertex_data)
        return cls(*center)

    def get_vertex(self):
        """
        Get a vertex from the data.

        Returns
        -------
        :class:`.Vertex`
            A vertex initialized from the data.

        Raises
        ------
        :class:`NotImplementedError`
            This is a virtual method which needs to be implemented in
            a subclass. The implementation should return an instance of
            the matching :class:`.Vertex` subclass.

        """

        raise NotImplementedError()

    def clone(self, clear_edges=False):
        """
        Return a clone.

        Parameters
        ----------
        clear_edges : :class:`bool`, optional
            ``True`` if the clone should not be connected to any edges.

        Returns
        -------
        :class:`VertexData`
            The clone.

        """

        clone = self.__class__.__new__(self.__class__)
        clone.id = self.id
        clone.position = list(self.position)
        clone.edges = [] if clear_edges else list(self.edges)
        clone.cell = np.array(self.cell)
        return clone

    def __str__(self):
        position = self.position.tolist()
        cell_id = self.cell.tolist()
        cell = '' if cell_id == [0, 0, 0] else f', cell={cell_id}'
        return f'VertexData(id={self.id}, position={position}{cell})'

    def __repr__(self):
        return str(self)


class Vertex:
    """
    Represents a vertex in a :class:`.TopologyGraph`.

    Attributes
    ----------
    id : :class:`int`
        The id of the vertex. This should be its index in
        :attr:`TopologyGraph.vertices`.

    """

    def __init__(self, data):
        """
        Initialize a :class:`.Vertex`.

        Parameters
        ----------
        data : :class:`.VertexData`
            The vertex data.

        """

        self.id = data.id
        self._position = np.array(data.position)
        # Holds the ids of edges the Vertex is connected to.
        self._edge_ids = [edge_data.id for edge_data in data.edges]
        self._cell = np.array(data.cell)
        # This holds the ConstructedMolecule that the vertex is used
        # to construct.
        self._mol = None

    def apply_scale(self, scale):
        """
        Scale the position by `scale`.

        Parameters
        ----------
        scale : :class:`float` or :class:`list`of :class:`float`
            The value by which the position of the :class:`Vertex` is
            scaled. Can be a single number if all axes are scaled by
            the same amount or a :class:`list` of three numbers if
            each axis is scaled by a different value.

        Returns
        -------
        :class:`Vertex`
            The vertex is returned.

        """

        self._position *= scale
        return self

    def clone(self, clear_edges=False):
        """
        Return a clone.

        Parameters
        ----------
        clear_edges : :class:`bool`, optional
            ``True`` if the clone should not be connected to any edges.

        Returns
        -------
        :class:`Vertex`
            The clone.

        """

        clone = self.__class__.__new__(self.__class__)
        clone.id = self.id
        clone._position = np.array(self._position)
        clone._cell = np.array(self._cell)
        clone._edge_ids = [] if clear_edges else list(self._edge_ids)
        clone._mol = self._mol
        return clone

    def get_position(self):
        """
        Return the position.

        Returns
        -------
        :class:`numpy.ndarray`
            The position of the :class:`Vertex`.

        """

        return np.array(self._position)

    def get_num_edges(self):
        """
        Return the number of connceted edge.

        Returns
        -------
        :class:`int`
            The number of connected edges.

        """

        return len(self._edge_ids)

    def get_edge_ids(self):
        """
        Yield the ids of connected edges.

        Yields
        ------
        :class:`int`
            The :class:`~.Edge.id` of a connected edge.

        """

        yield from self._edge_ids

    def get_cell(self):
        """
        Get the cell of the lattice in which the vertex is found.

        Returns
        -------
        :class:`numpy.ndarray`
            The cell of the lattice in which the vertex is found.

        """

        return np.array(self._cell)

    def set_contructed_molecule(self, mol):
        """
        Set the :class:`.ConstructedMolecule` being constructed.

        Parameters
        ----------
        mol : :class:`.ConstructedMolecule`
            The molecule being constructed.

        Returns
        -------
        :class:`.Vertex`
            The vertex.

        """

        self._mol = mol
        return self

    def place_building_block(self, building_block, vertices, edges):
        """
        Place `building_block` on the :class:`.Vertex`.

        Parameters
        ----------
        building_block : :class:`.Molecule`
            The building block molecule which is to be placed on the
            vertex.

        vertices : :class:`tuple` of :class:`.Vertex`
            All vertices in the topology graph. The index of each
            vertex must match its :class:`~.Vertex.id`.

        edges : :class:`tuple` of :class:`.Edge`
            All edges in the topology graph. The index of each
            edge must match its :class:`~.Edge.id`.

        Returns
        -------
        :class:`numpy.nadarray`
            The position matrix of `building_block` after being
            placed.

        Raises
        ------
        :class:`NotImplementedError`
            This is a virtual method, it needs to be implemented in a
            subclass.

        """

        raise NotImplementedError()

    def assign_func_groups_to_edges(
        self,
        building_block,
        vertices,
        edges
    ):
        """
        Assign functional groups to edges.

        Each :class:`.FunctionalGroup` of the `building_block` needs
        to be associated with one of the :class:`.Edge` instances in
        :attr:`edges`.

        Parameters
        ----------
        building_block : :class:`.Molecule`
            The building block molecule which is needs to have
            functional groups assigned to edges.

        vertices : :class:`tuple` of :class:`.Vertex`
            All vertices in the topology graph. The index of each
            vertex must match its :class:`~.Vertex.id`.

        edges : :class:`tuple` of :class:`.Edge`
            All edges in the topology graph. The index of each
            edge must match its :class:`~.Edge.id`.

        Returns
        -------
        :class:`dict`
            A mapping from the id of a functional group in
            `building_block` to the id of the edge in :attr:`edges` it
            is assigned to.

        Raises
        ------
        :class:`NotImplementedError`
            This is a virtual method, it needs to be implemented in a
            subclass.

        """

        raise NotImplementedError()

    def after_assign_func_groups_to_edges(
        self,
        building_block,
        func_groups,
        vertices,
        edges
    ):
        """
        Perform operations after functional groups have been assigned.

        This method is always executed serially. It is often useful
        when data needs to be transferred between vertices, which
        have been processed independently, in parallel.

        It does nothing by default, but should be overridden when
        necessary.

        Parameters
        ----------
        building_block : :class:`.Molecule`
            The building block molecule which is needs to have
            functional groups assigned to edges.

        func_groups : :class:`tuple` of :class:`.FunctionalGroup`
            The functional group clones added to the constructed
            molecule.

        vertices : :class:`tuple` of :class:`.Vertex`
            All vertices in the topology graph. The index of each
            vertex must match its :class:`~.Vertex.id`.

        edges : :class:`tuple` of :class:`.Edge`
            All edges in the topology graph. The index of each
            edge must match its :class:`~.Edge.id`.

        Returns
        -------
        None : :class:`NoneType`

        """

        return

    def _get_edge_centroid(self, centroid_edges, vertices):
        """
        Return the centroid of `centroid_edges`.

        Parameters
        ----------
        centroid_edges : :class:`iterable` of :class:`.Edge`
            The edges which are used to calculate the centroid.

        vertices : :class:`tuple` of :class:`.Vertex`
            All the vertices in the topology graph. Index of each
            vertex must be equal to its :class:`~.Vertex.id`.

        Returns
        -------
        :class:`numpy.ndarray`
            The centroid of the edges.

        """

        edge_positions = []
        for i, edge in enumerate(centroid_edges, 1):
            edge_positions.append(edge.get_position(self, vertices))
        return np.sum(edge_positions, axis=0) / i

    def _get_edge_plane_normal(self, reference, plane_edges, vertices):
        """
        Get the normal to the plane on which `edges` lie.

        Parameters
        ----------
        reference : :class:`numpy.ndarray`
            A reference direction vector. The direction of the returned
            normal is set such that its angle with with `reference`
            is always acute.

        plane_edges : :class:`iterable` of :class:`.Edge`
            The edges which are used to calculate the plane.
            If there are more than three, a plane of best fit across
            `edges` is returned.

        vertices : :class:`tuple` of :class:`.Vertex`
            All the vertices in the topology graph. Index of each
            vertex must be equal to its :class:`~.Vertex.id`.

        Returns
        -------
        :class:`numpy.ndarray`
            A unit vector which describes the normal to the plane of
            the edges.

        """

        edge_positions = []
        for i, edge in enumerate(plane_edges, 1):
            edge_positions.append(edge.get_position(self, vertices))
        edge_positions = np.array(edge_positions)

        centroid = np.sum(edge_positions, axis=0) / i
        normal = np.linalg.svd(edge_positions - centroid)[-1][2, :]

        if vector_angle(normal, reference) > np.pi/2:
            normal *= -1
        return normal

    def _get_molecule_centroid(self, atom_ids=None):
        """
        Get the centroid of the molecule being constructed.

        During construction :meth:`.Molecule.get_centroid` cannot be
        used, because the molecule is not fully constructed yet. This
        method acts as its replacement during construction.

        Parameters
        ----------
        atom_ids : :class:`iterable` of :class:`int`, optional
            The ids of atoms which are used to calculate the
            centroid. If ``None``, then all atoms will be used.

        Returns
        -------
        :class:`numpy.ndarray`
            The centroid of atoms specified by `atom_ids`.

        """

        if atom_ids is None:
            atom_ids = range(len(self._mol.atoms))
        elif not isinstance(atom_ids, (list, tuple)):
            atom_ids = list(atom_ids)

        return np.divide(
            np.sum(
                np.array(self._mol._position_matrix)[atom_ids, :],
                axis=0
            ),
            len(atom_ids)
        )

    def __str__(self):
        position = self._position.tolist()
        cell_id = self._cell.tolist()
        cell = '' if cell_id == [0, 0, 0] else f', cell={cell_id}'
        return f'Vertex(id={self.id}, position={position}{cell})'

    def __repr__(self):
        return str(self)


class EdgeData:
    """
    Holds data used to initialize a :class:`.Edge`.

    Attributes
    ----------
    id : :class:`int`
        The id of the edge. This is equal the index of the edge in
        :attr:`.TopologyGraph.edges`.

    vertices : :class:`list` of :class:`.VertexData`
        The vertices connected to the edge.

    periodicity : :class:`numpy.ndarray`
        The periodicity of the edge. For example, if ``(0, 0, 0)``
        then the edge is not periodic. If, ``(1, 0, -1)`` then the
        edge is periodic across the x axis in the positive
        direction, is not periodic across the y axis and is
        periodic across the z axis in the negative direction.

    lattice_constants : :class:`tuple` of :class:`numpy.ndarray`
        The a, b and c lattice constants as vectors in Cartesian
        coordinates.

    position : :class:`numpy.ndarray`
        The position of the edge.

    custom_position : :class:`bool`
        ``True`` if the :attr:`position` of the edge was set manually.
        ``False``if the position of the edge is the centroid of the
        connected :attr:`vertices`.

    """

    def __init__(
        self,
        *vertex_data,
        position=None,
        periodicity=None,
        lattice_constants=None
    ):
        """
        Initialize an :class:`EdgeData` instance.

        Parameters
        ----------
        *vertex_data : :class:`.VertexData`
            The vertices which the :class:`Edge` connects.

        position : :class:`numpy.ndarray`, optional
            The position of the edge. If ``None``, the centroid
            of `vertex_data` is used.

        periodicity : :class:`tuple` of :class:`int`, optional
            The periodicity of the edge. For example, if ``(0, 0, 0)``
            then the edge is not periodic. If, ``(1, 0, -1)`` then the
            edge is periodic across the x axis in the positive
            direction, is not periodic across the y axis and is
            periodic across the z axis in the negative direction. If
            ``None`` then the edge is not periodic.

        lattice_constants : :class:`iterable`, optional
            If the edge is periodic, the a, b and c lattice
            constants should be provided as vectors in Cartesian
            coordinates.

        """

        if periodicity is None:
            periodicity = [0, 0, 0]
        if lattice_constants is None:
            lattice_constants = ([0, 0, 0] for i in range(3))

        # This is set by TopologyGraph.__init__.
        self.id = None
        self.vertices = vertex_data
        self.periodicity = np.array(periodicity)
        self.custom_position = position is not None
        self.position = position
        self.lattice_constants = tuple(
            np.array(constant) for constant in lattice_constants
        )

        _position = 0
        for i, vertex in enumerate(vertex_data, 1):
            vertex.edges.append(self)

            if not self.custom_position:
                _position += vertex.position

        if not self.custom_position:
            self.position = _position / i

    def clone(
        self,
        vertex_map=None,
        recalculate_position=False,
        add_to_vertices=True
    ):
        """
        Return a clone.

        Parameters
        ----------
        vertex_map : :class:`dict`
            If the clone should hold different :class:`.VertexData`
            instances, then a :class:`dict` should be provided, which
            maps vertex data in the current :class:`.EdgeData` to the
            vertex data instances which should be used in the clone.
            Only vertex data instances which need to be changed need
            to be present in the `vertex_map`.

        recalculate_position : :class:`bool`, optional
            Toggle if the position of the clone should be recalculated
            from the vertices it connects or if it should inherit
            the position of the original edge.

        add_to_vertices : :class:`bool`, optional
            Toggles if the clone should be added to
            :attr:`.VertexData.edges`.

        Returns
        -------
        :class:`EdgeData`
            The clone.

        """

        clone = self.__class__.__new__(self.__class__)
        clone.id = self.id
        clone.vertices = tuple(
            vertex_map.get(v, v) for v in self.vertices
        )
        clone.periodicity = np.array(self.periodicity)
        clone.custom_position = self.custom_position
        clone.lattice_constants = tuple(
            np.array(constant) for constant in self.lattice_constants
        )
        if recalculate_position:
            vertex_positions = (
                vertex.position for vertex in clone.vertices
            )
            clone.position = np.divide(
                sum(vertex_positions),
                len(clone.vertices)
            )
            self.custom_position = False
        else:
            clone.position = np.array(self.position)

        if add_to_vertices:
            for vertex in clone.vertices:
                vertex.edges.append(clone)

        return clone

    def get_edge(self):
        """
        Get an :class:`.Edge` from the data.

        Returns
        -------
        :class:`Edge`
            The edge.

        """

        return Edge(self)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        vertices = ', '.join(
            str(vertex.id) for vertex in self.vertices
        )
        id_ = '' if self.id is None else f', id={self.id}'
        if self.custom_position:
            position = f', position={self.position!r}'
        else:
            position = ''

        if any(i != 0 for i in self.periodicity):
            periodicity = f', periodicity={tuple(self.periodicity)!r}'
        else:
            periodicity = ''

        return f'EdgeData({vertices}{id_}{position}{periodicity})'


class Edge:
    """
    Represents an edge in a topology graph.

    Note that some methods of this class will behave differently
    before and after :meth:`finalize` is called. Methods will switch
    from returning :class:`.Vertex` objects to returning :class:`int`
    objects.

    Attributes
    ----------
    id : :class:`int`
        The id of the edge. Matches the index of the edge in
        :attr:`.TopologyGraph.edges`.

    """

    def __init__(self, data):
        """
        Initialize an :class:`Edge`.

        Parameters
        ----------
        data : :class:`.EdgeData`
            The edge data.

        """

        self._vertex_ids = [vertex.id for vertex in data.vertices]
        self.id = data.id
        self._periodicity = np.array(data.periodicity)
        # The FunctionalGroup instances which the edge connects.
        # These will belong to the molecules placed on the vertices
        # connected by the edge.
        self._func_groups = []

        self._custom_position = data.custom_position
        self._position = np.array(data.position)
        self._lattice_constants = tuple(
            np.array(constant) for constant in data.lattice_constants
        )

    def get_periodicity(self):
        """
        Get the periodicity of the edge.

        Returns
        -------
        :class:`numpy.ndarray`
            The periodicity of the edge. If ``[0, 0, 0]`` the edge is
            not periodic, if ``[1, 0, -1]`` the edge is periodic going
            in the postive direction along the x axis, is not periodic
            across the y axis and is periodic in the negative direction
            along the z axis.

        """

        return np.array(self._periodicity)

    def is_periodic(self):
        """
        Return ``True`` if periodic.

        Returns
        -------
        :class:`bool`
            ``True`` if periodic.

        """

        return any(i != 0 for i in self._periodicity)

    def apply_scale(self, scale):
        """
        Scale the position by `scale`.

        Parameters
        ----------
        scale : :class:`float` or :class:`list`of :class:`float`
            The value by which the position of
            the :class:`Edge` is scaled. Can be a single number if all
            axes are scaled by the same amount or a :class:`list` of
            three numbers if each axis is scaled by a different value.

        Returns
        -------
        :class:`Edge`
            The edge is returned.


        """

        self._position *= scale
        self._lattice_constants = tuple(
            scale*constant for constant in self._lattice_constants
        )
        return self

    def clone(self):
        """
        Return a clone.

        Returns
        -------
        :class:`Edge`
            The clone.

        """

        clone = self.__class__.__new__(self.__class__)
        clone.id = self.id
        clone._func_groups = list(self._func_groups)
        clone._custom_position = self._custom_position
        clone._periodicity = np.array(self._periodicity)
        clone._lattice_constants = tuple(
            np.array(constant) for constant in self._lattice_constants
        )
        clone._vertex_ids = tuple(self._vertex_ids)
        clone._position = np.array(self._position)
        return clone

    def get_func_groups(self):
        """
        Get the functional groups connected by this edge.

        Returns
        -------
        :class:`tuple` of :class:`.FunctionalGroup`
            The functional groups connected by the edge.

        """

        return tuple(self._func_groups)

    def get_vertex_ids(self):
        """
        Get the connected vertices.

        Yields
        ------
        :class:`int`
            The id of a connected vertex.

        """

        yield from self._vertex_ids

    def assign_func_group(self, func_group):
        """
        Assign `func_group` to be connected by this edge.

        Parameters
        ----------
        func_group : :class:`.FunctionalGroup`
            The functional group to be assigned to the edge.

        Returns
        -------
        :class:`Edge`
            The edge is returned.

        """

        self._func_groups.append(func_group)

    def get_position(self, reference=None, vertices=None):
        """
        Return the position.

        Parameters
        ----------
        reference : :class:`.Vertex`, optional
            If the edge is periodic, the position returned will
            depend on which vertex the edge position is calculated
            relative to.

        vertices : :class:`tuple` of :class:`.Vertex`, optional
            All the vertices in the topology graph. Index of each
            vertex must be equal to its :class:`~.Vertex.id`. Only
            needs to be supplied if `reference` is supplied.

        Returns
        -------
        :class:`numpy.ndarray`
            The position of the :class:`Edge`.

        """

        if reference is None or not self.is_periodic():
            return np.array(self._position)

        other = vertices[
            next(v for v in self._vertex_ids if v != reference.id)
        ]
        direction = (
            1 if reference is vertices[self._vertex_ids[0]] else -1
        )
        end_cell = reference.get_cell() + direction*self._periodicity
        cell_shift = end_cell - other.get_cell()
        shift = 0
        for dim, constant in zip(cell_shift, self._lattice_constants):
            shift += dim*constant
        return (
            (other.get_position()+shift+reference.get_position()) / 2
        )

    def __str__(self):
        return repr(self)

    def __repr__(self):
        vertices = ', '.join(str(id_) for id_ in self._vertex_ids)
        if self._custom_position:
            position = f', position={self._position!r}'
        else:
            position = ''

        if any(i != 0 for i in self._periodicity):
            periodicity = f', periodicity={tuple(self._periodicity)!r}'
        else:
            periodicity = ''

        return f'Edge({vertices}{position}{periodicity})'


PlacementResult = namedtuple(
    'PlacementResult',
    ['building_block', 'vertex', 'assignments']
)


def _place_building_blocks(vertices, edges):

    def inner(vertex, building_block):
        vertex.place_building_block(building_block, vertices, edges)
        assignments = vertex.assign_func_groups_to_edges(
            building_block=building_block,
            vertices=vertices,
            edges=edges
        )
        return PlacementResult(building_block, vertex, assignments)

    return inner


class TopologyGraph:
    """
    Represents topology graphs of :class:`.ConstructedMolecule`.

    The topology graph is an abstract representation of a constructed
    molecule. The vertices indicate where building blocks are placed
    and the edges indicate which building blocks have bonds formed
    between them by the construction process.

    Vertices are responsible for placing the building block molecules.
    By initializing the vertices with different settings, they can
    position the building block molecules differently and therefore
    allow the user to easily specify a different structural isomer.

    Once a building block is placed on a vertex, the functional groups
    on the building block must be assigned to the different edges
    connected to the vertex. The number of functional groups in the
    building block must match the number of edges connected to the
    vertex.

    Once the functional groups are assigned to edges, each edge
    represents a reaction between the functional groups assigned to it.
    Note that an edge can be assigned more than two functional groups,
    in case you are dealing with something really exotic. The
    functional groups are then matched to an appropriate reaction,
    which generally creates bonds between the atoms of the functional
    groups. After this you will end up with a
    :class:`.ConstructedMolecule`.

    Attributes
    ----------
    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    """

    def __init__(
        self,
        vertex_data,
        edge_data,
        construction_stages,
        num_processes
    ):
        """
        Initialize an instance of :class:`.TopologyGraph`.

        Parameters
        ----------
        vertices : :class:`tuple` of :class:`.VertexData`
            The vertices which make up the graph.

        edges : :class:`tuple` of :class:`.EdgeData`
            The edges which make up the graph.

        construction_stages : :class:`tuple` of :class:`callable`
            A collection of callables, each of which takes a
            :class:`.Vertex` and returns ``True`` or ``False``.
            If the first :class:`callable` is applied to a  vertex in
            `vertices`, that vertex is is part of the first
            construction stage. The second :class:`callable` is then
            applied to all vertices not in the first stage and those
            which return ``True`` belong to the second stage and
            so on.

            Vertices which belong to the same construction stage
            all place building blocks together in parallel, before
            placement is done by any vertices which are part of a later
            stage. This breaks down parallel construction into
            serial stages if synchronization between stages is needed.

            If the topology graph is performing construction serially,
            then all vertices which belong to an earlier stage will
            place their building block before those at a later stage.

        num_processes : :class:`int`
            The number of parallel processes to create during
            :meth:`construct`.

        """

        self._set_data_ids(vertex_data)
        self._set_data_ids(edge_data)
        self.vertices = tuple(
            data.get_vertex() for data in vertex_data
        )
        self.edges = tuple(
            data.get_edge() for data in edge_data
        )
        self._construction_stages = construction_stages
        self._set_stages()
        self._num_processes = num_processes

    def _set_data_ids(self, data):
        for i, data in enumerate(data):
            data.id = i

    def _set_stages(self):
        self._stages = tuple(
            [] for i in range(len(self._construction_stages)+1)
        )
        for vertex in self.vertices:
            placed = False
            for i, stage in enumerate(self._construction_stages):
                if stage(vertex):
                    self._stages[i].append(vertex)
                    placed = True
                    break
            if not placed:
                self._stages[-1].append(vertex)

    def construct(self, mol):
        """
        Construct a :class:`.ConstructedMolecule`.

        Parameters
        ----------
        mol : :class:`.ConstructedMolecule`
            The :class:`.ConstructedMolecule` instance which needs to
            be constructed.

        Returns
        -------
        None : :class:`NoneType`

        """

        scale = self._get_scale(mol)
        vertices = tuple(self._get_vertex_clones(mol, scale))
        edges = tuple(self._get_edge_clones(scale))

        self._prepare(mol)
        self._place_building_blocks(mol, vertices, edges)

        vertices, edges = self._before_react(mol, vertices, edges)
        reactor = Reactor(mol)
        for edge in edges:
            reactor.add_reaction(
                func_groups=edge.get_func_groups(),
                periodicity=tuple(edge.get_periodicity())
            )
        reactor.finalize()

        self._clean_up(mol)

    def assign_building_blocks_to_vertices(self, building_blocks):
        """
        Assign `building_blocks` to :attr:`vertices`.

        Parameters
        ----------
        building_blocks : :class:`list` of :class:`.Molecule`
            The :class:`.BuildingBlock` and
            :class:`ConstructedMolecule` instances which
            represent the building block molecules used for
            construction. Only one instance is present per building
            block molecule, even if multiples of that building block
            join up to form the :class:`ConstructedMolecule`.

        Returns
        -------
        :class:`dict`
            Maps the `building_blocks`, to the
            :class:`~.topologies.base.Vertex` objects in
            :attr:`vertices` they are placed on during construction.
            The :class:`dict` has the form

            .. code-block:: python

                building_block_vertices = {
                    BuildingBlock(...): [Vertex(...), Vertex(...)],
                    BuildingBlock(...): [
                        Vertex(...),
                        Vertex(...),
                        Vertex(...),
                    ]
                    ConstructedMolecule(...): [Vertex(...)]
                }

        Raises
        ------
        :class:`NotImplementedError`
            This is a virtual method which needs to be implemented in
            a subclass.

        """

        raise NotImplementedError()

    def _get_scale(self, mol):
        """
        Get the scale used for vertex and edge positions.

        Parameters
        ----------
        mol : :class:`.ConstructedMolecule`
            The molecule being constructed.

        Returns
        -------
        :class:`float` or :class:`list` of :class:`float`
            The value by which the position of each :class:`Vertex` and
            is :class:`Edge` is scaled. Can be a single number if all
            axes are scaled by the same amount or a :class:`list` of
            three numbers if each axis is scaled by a different value.

        Raises
        ------
        :class:`NotImplementedError`
            This is a virtual method and needs to be implemented in a
            subclass.

        """

        raise NotImplementedError()

    def _get_vertex_clones(self, mol, scale):
        """
        Yield clones of :attr:`vertices`.

        The order of yielded clones corresponds to the order in
        :attr:`vertices`

        Notes
        -----
        Clones are necessary so that multiple :meth:`construct`
        calls can be done asynchronously and so that the state of the
        original :class:`.Vertex` objects is not
        changed by the construction process.

        Parameters
        ----------
        mol : :class:`.ConstructedMolecule`
            The molecule being constructed.

        scale : :class:`float` or :class:`list` of :class:`float`
            The value by which the position of each :class:`Vertex` is
            scaled. Can be a single number if all axes are scaled by
            the same amount or a :class:`list` of three numbers if
            each axis is scaled by a different value.

        Yields
        -------
        :class:`.Vertex`
            A vertex clone.

        """

        for vertex in self.vertices:
            yield (
                vertex
                .clone()
                .set_contructed_molecule(mol)
                .apply_scale(scale)
            )

    def _get_edge_clones(self, scale):
        """
        Yield clones of :attr:`edges`.

        The order of yielded edges corresponds to the order in
        :attr:`edges`.

        Parameters
        ----------
        scale : :class:`float` or :class:`list` of :class:`float`
            The value by which the position of each :class:`Edge` is
            scaled. Can be a single number if all axes are scaled by
            the same amount or a :class:`list` of three numbers if
            each axis is scaled by a different value.

        Yields
        -------
        :class:`.Edge`
            An edge clone.

        """

        for edge in self.edges:
            yield edge.clone().apply_scale(scale)

    def _before_react(self, mol, vertices, edges):
        return vertices, edges

    def _prepare(self, mol):
        """
        Do preprocessing on `mol` before construction.

        Parameters
        ----------
        mol : :class:`.ConstructedMolecule`
            The molecule being constructed.

        Returns
        -------
        None : :class:`NoneType`

        """

        return

    def _place_building_blocks(self, mol, vertices, edges):
        """
        Place building blocks in `mol` on :attr:`vertices`.

        Parameters
        ----------
        mol : :class:`.ConstructedMolecule`
            The molecule being constructed.

        vertices : :class:`tuple` of :class:`.Vertex`
            The vertex clones used for construction.

        edges : :class:`tuple` of :class:`.Edge`
            The edge clones used for construction.

        Returns
        -------
        None : :class:`NoneType`

        """

        if self._num_processes == 1:
            return self._place_building_blocks_serial(
                mol=mol,
                vertices=vertices,
                edges=edges
            )
        else:
            return self._place_building_blocks_parallel(
                mol=mol,
                vertices=vertices,
                edges=edges
            )

    def _get_atom_map(self, mol, bb, bb_id):
        atom_map = {}
        for atom in bb.atoms:
            atom_clone = atom.clone()
            atom_clone.id = len(mol.atoms)
            atom_clone.building_block = bb
            atom_clone.building_block_id = bb_id
            atom_map[atom] = atom_clone
            mol.atoms.append(atom_clone)
        return atom_map

    def _assign_func_groups_to_edges(
        self,
        mol,
        bb,
        bb_id,
        edges,
        assignments
    ):
        atom_map = self._get_atom_map(mol, bb, bb_id)
        mol.func_groups.extend(
            fg.clone(atom_map) for fg in bb.func_groups
        )
        num_fgs = len(bb.func_groups)
        for fg_id, edge_id in assignments.items():
            edges[edge_id].assign_func_group(
                func_group=mol.func_groups[-num_fgs+fg_id]
            )
        return atom_map

    def _place_building_blocks_serial(self, mol, vertices, edges):
        bb_id = 0

        vertex_building_blocks = {
            vertex: bb
            for bb, vertices in mol.building_block_vertices.items()
            for vertex in vertices
        }
        # Use a shorter alias.
        counter = mol.building_block_counter
        for stage in self._stages:
            for instance_vertex in stage:
                vertex = vertices[instance_vertex.id]
                bb = vertex_building_blocks[instance_vertex]
                original_coords = bb.get_position_matrix()

                mol._position_matrix.extend(
                    vertex.place_building_block(bb, vertices, edges)
                )
                assignments = vertex.assign_func_groups_to_edges(
                    building_block=bb,
                    vertices=vertices,
                    edges=edges
                )
                atom_map = self._assign_func_groups_to_edges(
                    mol=mol,
                    bb=bb,
                    bb_id=bb_id,
                    edges=edges,
                    assignments=assignments
                )
                # Perform additional, miscellaneous operations.
                vertex.after_assign_func_groups_to_edges(
                    building_block=bb,
                    func_groups=mol.func_groups[-len(bb.func_groups):],
                    vertices=vertices,
                    edges=edges
                )

                bb.set_position_matrix(original_coords)
                mol.bonds.extend(b.clone(atom_map) for b in bb.bonds)
                counter.update([bb])
                bb_id += 1

    def _place_building_blocks_parallel(self, mol, vertices, edges):
        bb_id = 0

        vertex_building_blocks = {
            vertex: bb
            for bb, vertices in mol.building_block_vertices.items()
            for vertex in vertices
        }
        bb_map = {
            bb.get_identity_key(): bb
            for bb in mol.get_building_blocks()
        }
        # Use a shorter alias.
        counter = mol.building_block_counter
        place = _place_building_blocks(vertices, edges)
        with pathos.pools.ProcessPool(self._num_processes) as pool:
            for stage in self._stages:
                verts = []
                bbs = []
                for instance_vertex in stage:
                    verts.append(vertices[instance_vertex.id])
                    bbs.append(vertex_building_blocks[instance_vertex])
                results = pool.map(place, verts, bbs)

                for result in results:
                    result_bb = result.building_block
                    bb = bb_map[result_bb.get_identity_key()]

                    mol._position_matrix.extend(
                        result_bb.get_position_matrix()
                    )
                    atom_map = self._assign_func_groups_to_edges(
                        mol=mol,
                        bb=bb,
                        bb_id=bb_id,
                        edges=edges,
                        assignments=result.assignments
                    )

                    # Perform additional, miscellaneous operations.
                    vertex = vertices[result.vertex.id]
                    num_fgs = len(bb.func_groups)
                    vertex.after_assign_func_groups_to_edges(
                        building_block=result_bb,
                        func_groups=mol.func_groups[-num_fgs:],
                        vertices=vertices,
                        edges=edges
                    )
                    mol.bonds.extend(
                        b.clone(atom_map) for b in bb.bonds
                    )
                    counter.update([bb])
                    bb_id += 1

    def _clean_up(self, mol):
        mol._position_matrix = np.array(mol._position_matrix).T
        for i, atom in enumerate(mol.atoms):
            atom.id = i

    def __str__(self):
        return repr(self)

    def __repr__(self):
        raise NotImplementedError()
